import shutil
import subprocess
import sys
import tarfile
from pathlib import Path
from unittest import mock

import pytest

from briefcase.console import Console, Log
from briefcase.exceptions import BriefcaseCommandError
from briefcase.platforms.linux import system
from briefcase.platforms.linux.system import LinuxSystemPackageCommand

from ....utils import create_file, create_tgz_file


@pytest.fixture
def package_command(first_app, tmp_path):
    command = LinuxSystemPackageCommand(
        logger=Log(),
        console=Console(),
        base_path=tmp_path / "base_path",
        data_path=tmp_path / "briefcase",
    )
    command.tools.home_path = tmp_path / "home"

    # Mock ABI from packaging system
    command._pkg_abi = "wonky"

    # Mock the app context
    command.tools.app_tools[first_app].app_context = mock.MagicMock()

    # Mock shutil
    command.tools.shutil = mock.MagicMock()

    # Make the mock copy still copy
    command.tools.shutil.copy = mock.MagicMock(side_effect=shutil.copy)

    # Make the mock make_archive still package tarballs
    command.tools.shutil.make_archive = mock.MagicMock(side_effect=shutil.make_archive)

    # Make the mock rmtree still remove content
    command.tools.shutil.rmtree = mock.MagicMock(side_effect=shutil.rmtree)

    # Mock not using docker
    command.target_image = None
    command.extra_docker_build_args = []

    return command


@pytest.fixture
def first_app_pkg(first_app, tmp_path):
    # Mock an Arch app
    first_app.python_version_tag = "3.10"
    first_app.target_vendor_base = "arch"
    first_app.packaging_format = "pkg"
    first_app.glibc_version = "2.99"
    first_app.description = "Description for the app"

    # Mock the side effects of building the app
    usr_dir = (
        tmp_path / "base_path/build/first-app/somevendor/surprising/first-app-0.0.1/usr"
    )

    # Create the binary
    create_file(usr_dir / "bin/first-app", "binary")

    # Files in an app-named folder
    create_file(usr_dir / "share/doc/first-app/license", "license")
    create_file(usr_dir / "share/doc/first-app/UserManual", "manual")

    # A share file in an app-named folder
    create_file(usr_dir / "share/man/man1/first-app.1.gz", "man")

    return first_app


def test_verify_no_docker(package_command, first_app_pkg, monkeypatch):
    """If not using docker, existence of makepkg is verified."""
    # Mock the existence of makepkg
    package_command.tools.shutil.which = mock.MagicMock(return_value="/mybin/makepkg")

    # App tools can be verified
    package_command.verify_app_tools(first_app_pkg)

    # makepkg was inspected
    package_command.tools.shutil.which.assert_called_once_with("makepkg")


@pytest.mark.parametrize(
    "vendor_base, error_msg",
    [
        (
            "arch",
            "Can't find the makepkg tools. Try running `sudo pacman -Syu pacman`.",
        ),
        (None, "Can't find the makepkg tool. Install this first to package the pkg."),
    ],
)
def test_verify_makepkg_missing(
    package_command,
    first_app_pkg,
    vendor_base,
    error_msg,
    monkeypatch,
):
    """If makepkg isn't installed, an error is raised."""
    # Mock distro so packager is found or not appropriately
    first_app_pkg.target_vendor_base = vendor_base

    # Mock packager as missing
    package_command.tools.shutil.which = mock.MagicMock(return_value="")

    # Verifying app tools will raise an error
    with pytest.raises(BriefcaseCommandError, match=error_msg):
        package_command.verify_app_tools(first_app_pkg)

    # which was called for makepkg
    package_command.tools.shutil.which.assert_called_once_with("makepkg")


def test_verify_docker(package_command, first_app_pkg, monkeypatch):
    """If using Docker, no tool checks are needed."""
    # Mock using docker
    package_command.target_image = "somevendor:surprising"

    # Mock the existence of a valid non-docker system Python
    # with the same major/minor as the current Python,
    # plus the existence of makepkg
    python3 = mock.MagicMock()
    python3.resolve.return_value = Path(
        f"/usr/bin/python{sys.version_info.major}.{sys.version_info.minor}"
    )

    makepkg = mock.MagicMock()
    makepkg.exists.return_value = False

    mock_Path = mock.MagicMock(side_effect=[python3, makepkg])
    monkeypatch.setattr(system, "Path", mock_Path)

    # App tools can be verified
    package_command.verify_app_tools(first_app_pkg)

    # makepkg was not inspected
    makepkg.exists.assert_not_called()


@pytest.mark.skipif(sys.platform == "win32", reason="Can't build PKGs on Windows")
def test_pkg_package(package_command, first_app_pkg, tmp_path):
    """A pkg app can be packaged."""
    bundle_path = tmp_path / "base_path/build/first-app/somevendor/surprising"

    # Package the app
    package_command.package_app(first_app_pkg)

    # The CHANGELOG file is copied
    assert (bundle_path / "pkgbuild/CHANGELOG").exists()

    # The PKGBUILD file is written
    assert (bundle_path / "pkgbuild/PKGBUILD").exists()
    with (bundle_path / "pkgbuild/PKGBUILD").open(encoding="utf-8") as f:
        assert f.read() == "\n".join(
            [
                "# Maintainer: Megacorp <maintainer@example.com>",
                'export PACKAGER="Megacorp <maintainer@example.com>"',
                "pkgname=first-app",
                "pkgver=0.0.1",
                "pkgrel=1",
                'pkgdesc="Description for the app"',
                "arch=('wonky')",
                'url="https://example.com/first-app"',
                "license=('Unknown')",
                "depends=('glibc>=2.99' 'python3')",
                "changelog=CHANGELOG",
                'source=("$pkgname-$pkgver.tar.gz")',
                "md5sums=('SKIP')",
                "options=('!strip')",
                "package() {",
                '    cp -r "$srcdir/$pkgname-$pkgver/usr/" "$pkgdir"/usr/',
                "}",
            ]
        )

    # A source tarball was created with the right content
    archive_file = bundle_path / "pkgbuild/first-app-0.0.1.tar.gz"
    assert archive_file.exists()
    with tarfile.open(archive_file, "r:gz") as archive:
        assert sorted(archive.getnames()) == [
            "first-app-0.0.1",
            "first-app-0.0.1/usr",
            "first-app-0.0.1/usr/bin",
            "first-app-0.0.1/usr/bin/first-app",
            "first-app-0.0.1/usr/lib",
            "first-app-0.0.1/usr/lib/first-app",
            "first-app-0.0.1/usr/lib/first-app/app",
            "first-app-0.0.1/usr/lib/first-app/app/support.so",
            "first-app-0.0.1/usr/lib/first-app/app/support_same_perms.so",
            "first-app-0.0.1/usr/lib/first-app/app_packages",
            "first-app-0.0.1/usr/lib/first-app/app_packages/firstlib",
            "first-app-0.0.1/usr/lib/first-app/app_packages/firstlib/first.so",
            "first-app-0.0.1/usr/lib/first-app/app_packages/firstlib/first.so.1.0",
            "first-app-0.0.1/usr/lib/first-app/app_packages/secondlib",
            "first-app-0.0.1/usr/lib/first-app/app_packages/secondlib/second_a.so",
            "first-app-0.0.1/usr/lib/first-app/app_packages/secondlib/second_b.so",
            "first-app-0.0.1/usr/share",
            "first-app-0.0.1/usr/share/doc",
            "first-app-0.0.1/usr/share/doc/first-app",
            "first-app-0.0.1/usr/share/doc/first-app/UserManual",
            "first-app-0.0.1/usr/share/doc/first-app/license",
            "first-app-0.0.1/usr/share/man",
            "first-app-0.0.1/usr/share/man/man1",
            "first-app-0.0.1/usr/share/man/man1/first-app.1.gz",
        ]

    # makepkg was invoked
    package_command.tools.app_tools[
        first_app_pkg
    ].app_context.run.assert_called_once_with(
        [
            "makepkg",
        ],
        check=True,
        cwd=(bundle_path / "pkgbuild"),
        env={"PKGEXT": ".pkg.tar.zst"},
    )

    # The pkg was moved into the final location
    package_command.tools.shutil.move.assert_called_once_with(
        bundle_path / "pkgbuild/first-app-0.0.1-1-wonky.pkg.tar.zst",
        tmp_path / "base_path/dist/first-app-0.0.1-1-wonky.pkg.tar.zst",
    )


@pytest.mark.skipif(sys.platform == "win32", reason="Can't build PKGs on Windows")
def test_pkg_re_package(package_command, first_app_pkg, tmp_path):
    """A pkg app that has previously been packaged can be re-packaged."""
    bundle_path = tmp_path / "base_path/build/first-app/somevendor/surprising"

    # Create an old PKGBUILD file and tarball that will be overwritten.
    create_file(
        bundle_path / "pkgbuild/PKGBUILD",
        "Old PKGBUILD content",
    )
    create_tgz_file(
        bundle_path / "pkgbuild/first-app-0.0.1.tar.gz",
        [("old.txt", "old content")],
    )

    # Package the app
    package_command.package_app(first_app_pkg)

    # The CHANGELOG file is copied
    assert (bundle_path / "pkgbuild/CHANGELOG").exists()

    # The PKGBUILD file is written
    assert (bundle_path / "pkgbuild/PKGBUILD").exists()
    with (bundle_path / "pkgbuild/PKGBUILD").open(encoding="utf-8") as f:
        assert f.read() == "\n".join(
            [
                "# Maintainer: Megacorp <maintainer@example.com>",
                'export PACKAGER="Megacorp <maintainer@example.com>"',
                "pkgname=first-app",
                "pkgver=0.0.1",
                "pkgrel=1",
                'pkgdesc="Description for the app"',
                "arch=('wonky')",
                'url="https://example.com/first-app"',
                "license=('Unknown')",
                "depends=('glibc>=2.99' 'python3')",
                "changelog=CHANGELOG",
                'source=("$pkgname-$pkgver.tar.gz")',
                "md5sums=('SKIP')",
                "options=('!strip')",
                "package() {",
                '    cp -r "$srcdir/$pkgname-$pkgver/usr/" "$pkgdir"/usr/',
                "}",
            ]
        )

    # A source tarball was created with the right content
    archive_file = bundle_path / "pkgbuild/first-app-0.0.1.tar.gz"
    assert archive_file.exists()
    with tarfile.open(archive_file, "r:gz") as archive:
        assert sorted(archive.getnames()) == [
            "first-app-0.0.1",
            "first-app-0.0.1/usr",
            "first-app-0.0.1/usr/bin",
            "first-app-0.0.1/usr/bin/first-app",
            "first-app-0.0.1/usr/lib",
            "first-app-0.0.1/usr/lib/first-app",
            "first-app-0.0.1/usr/lib/first-app/app",
            "first-app-0.0.1/usr/lib/first-app/app/support.so",
            "first-app-0.0.1/usr/lib/first-app/app/support_same_perms.so",
            "first-app-0.0.1/usr/lib/first-app/app_packages",
            "first-app-0.0.1/usr/lib/first-app/app_packages/firstlib",
            "first-app-0.0.1/usr/lib/first-app/app_packages/firstlib/first.so",
            "first-app-0.0.1/usr/lib/first-app/app_packages/firstlib/first.so.1.0",
            "first-app-0.0.1/usr/lib/first-app/app_packages/secondlib",
            "first-app-0.0.1/usr/lib/first-app/app_packages/secondlib/second_a.so",
            "first-app-0.0.1/usr/lib/first-app/app_packages/secondlib/second_b.so",
            "first-app-0.0.1/usr/share",
            "first-app-0.0.1/usr/share/doc",
            "first-app-0.0.1/usr/share/doc/first-app",
            "first-app-0.0.1/usr/share/doc/first-app/UserManual",
            "first-app-0.0.1/usr/share/doc/first-app/license",
            "first-app-0.0.1/usr/share/man",
            "first-app-0.0.1/usr/share/man/man1",
            "first-app-0.0.1/usr/share/man/man1/first-app.1.gz",
        ]

    # makepkg was invoked
    package_command.tools.app_tools[
        first_app_pkg
    ].app_context.run.assert_called_once_with(
        [
            "makepkg",
        ],
        check=True,
        cwd=(bundle_path / "pkgbuild"),
        env={"PKGEXT": ".pkg.tar.zst"},
    )

    # The pkg was moved into the final location
    package_command.tools.shutil.move.assert_called_once_with(
        bundle_path / "pkgbuild/first-app-0.0.1-1-wonky.pkg.tar.zst",
        tmp_path / "base_path/dist/first-app-0.0.1-1-wonky.pkg.tar.zst",
    )


@pytest.mark.skipif(sys.platform == "win32", reason="Can't build PKGs on Windows")
def test_pkg_package_no_description(package_command, first_app_pkg, tmp_path):
    """A pkg app without a description raises an error."""
    bundle_path = tmp_path / "base_path/build/first-app/somevendor/surprising"

    # Delete the long description
    first_app_pkg.description = None

    # Packaging the app will fail
    with pytest.raises(
        BriefcaseCommandError,
        match=r"App configuration does not define `description`. Arch projects require a description.",
    ):
        package_command.package_app(first_app_pkg)

    # The CHANGELOG file, PKGBUILD file and tarball won't be written
    assert not (bundle_path / "pkgbuild/CHANGELOG").exists()
    assert not (bundle_path / "pkgbuild/PKGBUILD").exists()
    assert not (bundle_path / "pkgbuild/first-app-0.0.1.tar.gz").exists()


@pytest.mark.skipif(sys.platform == "win32", reason="Can't build PKGs on Windows")
def test_pkg_package_extra_requirements(package_command, first_app_pkg, tmp_path):
    """A pkg app can be packaged with extra runtime requirements and config features."""
    bundle_path = tmp_path / "base_path/build/first-app/somevendor/surprising"

    # Add system requirements and other optional settings.
    first_app_pkg.system_runtime_requires = ["first", "second"]
    first_app_pkg.revision = 1

    # Package the app
    package_command.package_app(first_app_pkg)

    # The CHANGELOG file is copied
    assert (bundle_path / "pkgbuild/CHANGELOG").exists()

    # The PKGBUILD file is written
    assert (bundle_path / "pkgbuild/PKGBUILD").exists()
    with (bundle_path / "pkgbuild/PKGBUILD").open(encoding="utf-8") as f:
        assert f.read() == "\n".join(
            [
                "# Maintainer: Megacorp <maintainer@example.com>",
                'export PACKAGER="Megacorp <maintainer@example.com>"',
                "pkgname=first-app",
                "pkgver=0.0.1",
                "pkgrel=1",
                'pkgdesc="Description for the app"',
                "arch=('wonky')",
                'url="https://example.com/first-app"',
                "license=('Unknown')",
                "depends=('glibc>=2.99' 'python3' 'first' 'second')",
                "changelog=CHANGELOG",
                'source=("$pkgname-$pkgver.tar.gz")',
                "md5sums=('SKIP')",
                "options=('!strip')",
                "package() {",
                '    cp -r "$srcdir/$pkgname-$pkgver/usr/" "$pkgdir"/usr/',
                "}",
            ]
        )

    # A source tarball was created
    archive_file = bundle_path / "pkgbuild/first-app-0.0.1.tar.gz"
    assert archive_file.exists()

    # makepkg was invoked
    package_command.tools.app_tools[
        first_app_pkg
    ].app_context.run.assert_called_once_with(
        [
            "makepkg",
        ],
        check=True,
        cwd=(bundle_path / "pkgbuild"),
        env={"PKGEXT": ".pkg.tar.zst"},
    )

    # The pkg was moved into the final location
    package_command.tools.shutil.move.assert_called_once_with(
        bundle_path / "pkgbuild/first-app-0.0.1-1-wonky.pkg.tar.zst",
        tmp_path / "base_path/dist/first-app-0.0.1-1-wonky.pkg.tar.zst",
    )


@pytest.mark.skipif(sys.platform == "win32", reason="Can't build PKGs on Windows")
def test_pkg_package_failure(package_command, first_app_pkg, tmp_path):
    """If a packaging doesn't succeed, an error is raised."""
    bundle_path = tmp_path / "base_path/build/first-app/somevendor/surprising"

    # Mock a packaging failure
    package_command.tools.app_tools[first_app_pkg].app_context.run.side_effect = (
        subprocess.CalledProcessError(cmd="makepkg ...", returncode=-1)
    )

    # Package the app; this will fail
    with pytest.raises(
        BriefcaseCommandError,
        match=r"Error while building .pkg.tar.zst package for first-app.",
    ):
        package_command.package_app(first_app_pkg)

    # The CHANGELOG file is copied
    assert (bundle_path / "pkgbuild/CHANGELOG").exists()

    # The PKGBUILD file is written
    assert (bundle_path / "pkgbuild/PKGBUILD").exists()

    # A source tarball was created
    archive_file = bundle_path / "pkgbuild/first-app-0.0.1.tar.gz"
    assert archive_file.exists()

    # makepkg was invoked
    package_command.tools.app_tools[
        first_app_pkg
    ].app_context.run.assert_called_once_with(
        [
            "makepkg",
        ],
        check=True,
        cwd=(bundle_path / "pkgbuild"),
        env={"PKGEXT": ".pkg.tar.zst"},
    )

    # The pkg wasn't built, so it wasn't moved.
    package_command.tools.shutil.move.assert_not_called()


@pytest.mark.skipif(sys.platform == "win32", reason="Can't build PKGs on Windows")
def test_no_changelog(package_command, first_app_pkg, tmp_path):
    """If a packaging doesn't succeed, an error is raised."""
    bundle_path = tmp_path / "base_path/build/first-app/somevendor/surprising"

    # Remove the changelog file
    (tmp_path / "base_path/CHANGELOG").unlink()

    # Package the app; this will fail
    with pytest.raises(
        BriefcaseCommandError, match=r"Your project does not contain a CHANGELOG file."
    ):
        package_command.package_app(first_app_pkg)

    # The CHANGELOG file will not be copied
    assert not (bundle_path / "pkgbuild/CHANGELOG").exists()

    # The PKGBUILD file will not exist (as existence of changelog is checked before writing the PKGBUILD file)
    assert not (bundle_path / "pkgbuild/PKGBUILD").exists()

    # No source tarball was created
    archive_file = bundle_path / "pkgbuild/first-app-0.0.1.tar.gz"
    assert not archive_file.exists()

    # makepkg wasn't invoked
    package_command.tools.app_tools[first_app_pkg].app_context.run.assert_not_called()

    # The pkg wasn't built, so it wasn't moved.
    package_command.tools.shutil.move.assert_not_called()



from unittest.mock import MagicMock

import briefcase.platforms.linux.system
from briefcase.integrations.docker import Docker, DockerAppContext
from briefcase.integrations.subprocess import Subprocess


def test_linux_no_docker(create_command, first_app_config, monkeypatch):
    """If Docker is disabled on Linux, the app_context is Subprocess."""
    create_command.tools.host_os = "Linux"
    create_command.target_image = None

    # Force a dummy vendor:codename for test purposes.
    first_app_config.target_vendor = "somevendor"
    first_app_config.target_codename = "surprising"
    first_app_config.target_vendor_base = "basevendor"

    # Mock the existence of a valid non-docker system Python
    create_command.verify_system_python = MagicMock()

    # Verify the tools
    create_command.verify_tools()
    create_command.verify_app_tools(app=first_app_config)

    # No error and Subprocess is used.
    assert isinstance(create_command.tools[first_app_config].app_context, Subprocess)
    # Docker is not verified.
    assert not hasattr(create_command.tools, "docker")
    # System python is verified
    create_command.verify_system_python.assert_called_once_with()

    # Reset the mock, then invoke verify_app_tools a second time.
    create_command.verify_system_python.reset_mock()
    create_command.verify_app_tools(app=first_app_config)

    # Python will *not* be verified a second time.
    create_command.verify_system_python.assert_not_called()


def test_linux_docker(create_command, first_app_config, tmp_path, monkeypatch):
    """If Docker is enabled on Linux, the Docker alias is set."""
    create_command.tools.host_os = "Linux"
    create_command.target_image = "somevendor:surprising"
    create_command.extra_docker_build_args = ["--option-one", "--option-two"]

    # Force a dummy vendor:codename for test purposes.
    first_app_config.target_vendor = "somevendor"
    first_app_config.target_codename = "surprising"
    first_app_config.target_vendor_base = "basevendor"
    first_app_config.python_version_tag = "3"

    # Mock Docker tool verification
    mock__version_compat = MagicMock(spec=Docker._version_compat)
    mock__user_access = MagicMock(spec=Docker._user_access)
    mock__buildx_installed = MagicMock(spec=Docker._buildx_installed)
    mock__is_user_mapping_enabled = MagicMock(spec=Docker._is_user_mapping_enabled)
    monkeypatch.setattr(
        briefcase.platforms.linux.system.Docker,
        "_version_compat",
        mock__version_compat,
    )
    monkeypatch.setattr(
        briefcase.platforms.linux.system.Docker,
        "_user_access",
        mock__user_access,
    )
    monkeypatch.setattr(
        briefcase.platforms.linux.system.Docker,
        "_buildx_installed",
        mock__buildx_installed,
    )
    monkeypatch.setattr(
        briefcase.platforms.linux.system.Docker,
        "_is_user_mapping_enabled",
        mock__is_user_mapping_enabled,
    )
    mock_docker_app_context_verify = MagicMock(spec=DockerAppContext.verify)
    monkeypatch.setattr(
        briefcase.platforms.linux.system.DockerAppContext,
        "verify",
        mock_docker_app_context_verify,
    )
    create_command.verify_python = MagicMock()

    # Verify the tools
    create_command.verify_tools()
    create_command.verify_app_tools(app=first_app_config)

    # Docker and Docker app context are verified
    mock__version_compat.assert_called_with(tools=create_command.tools)
    mock__user_access.assert_called_with(tools=create_command.tools)
    mock__buildx_installed.assert_called_with(tools=create_command.tools)
    mock__is_user_mapping_enabled.assert_called_with("somevendor:surprising")
    assert isinstance(create_command.tools.docker, Docker)
    mock_docker_app_context_verify.assert_called_with(
        tools=create_command.tools,
        app=first_app_config,
        image_tag="briefcase/com.example.first-app:somevendor-surprising",
        dockerfile_path=tmp_path
        / "base_path/build/first-app/somevendor/surprising/Dockerfile",
        app_base_path=tmp_path / "base_path",
        host_bundle_path=tmp_path / "base_path/build/first-app/somevendor/surprising",
        host_data_path=tmp_path / "briefcase",
        python_version="3",
        extra_build_args=["--option-one", "--option-two"],
    )

    # Python was also verified
    create_command.verify_python.assert_called_once_with(first_app_config)

    # Reset the mock, then invoke verify_app_tools a second time.
    create_command.verify_python.reset_mock()
    create_command.verify_app_tools(app=first_app_config)

    # Python will *not* be verified a second time.
    create_command.verify_python.assert_not_called()


def test_non_linux_docker(create_command, first_app_config, tmp_path, monkeypatch):
    """If Docker is enabled on non-Linux, the Docker alias is set."""
    create_command.tools.host_os = "Darwin"
    create_command.target_image = "somevendor:surprising"
    create_command.extra_docker_build_args = ["--option-one", "--option-two"]

    # Force a dummy vendor:codename for test purposes.
    first_app_config.target_vendor = "somevendor"
    first_app_config.target_codename = "surprising"
    first_app_config.target_vendor_base = "basevendor"
    first_app_config.python_version_tag = "3"

    # Mock Docker tool verification
    mock__version_compat = MagicMock(spec=Docker._version_compat)
    mock__user_access = MagicMock(spec=Docker._user_access)
    mock__buildx_installed = MagicMock(spec=Docker._buildx_installed)
    mock__is_user_mapping_enabled = MagicMock(spec=Docker._is_user_mapping_enabled)
    monkeypatch.setattr(
        briefcase.platforms.linux.system.Docker,
        "_version_compat",
        mock__version_compat,
    )
    monkeypatch.setattr(
        briefcase.platforms.linux.system.Docker,
        "_user_access",
        mock__user_access,
    )
    monkeypatch.setattr(
        briefcase.platforms.linux.system.Docker,
        "_buildx_installed",
        mock__buildx_installed,
    )
    monkeypatch.setattr(
        briefcase.platforms.linux.system.Docker,
        "_is_user_mapping_enabled",
        mock__is_user_mapping_enabled,
    )
    mock_docker_app_context_verify = MagicMock(spec=DockerAppContext.verify)
    monkeypatch.setattr(
        briefcase.platforms.linux.system.DockerAppContext,
        "verify",
        mock_docker_app_context_verify,
    )
    create_command.verify_python = MagicMock()

    # Verify the tools
    create_command.verify_tools()
    create_command.verify_app_tools(app=first_app_config)

    # Docker and Docker app context are verified
    mock__version_compat.assert_called_with(tools=create_command.tools)
    mock__user_access.assert_called_with(tools=create_command.tools)
    mock__buildx_installed.assert_called_with(tools=create_command.tools)
    mock__is_user_mapping_enabled.assert_called_with("somevendor:surprising")
    assert isinstance(create_command.tools.docker, Docker)
    mock_docker_app_context_verify.assert_called_with(
        tools=create_command.tools,
        app=first_app_config,
        image_tag="briefcase/com.example.first-app:somevendor-surprising",
        dockerfile_path=tmp_path
        / "base_path/build/first-app/somevendor/surprising/Dockerfile",
        app_base_path=tmp_path / "base_path",
        host_bundle_path=tmp_path / "base_path/build/first-app/somevendor/surprising",
        host_data_path=tmp_path / "briefcase",
        python_version="3",
        extra_build_args=["--option-one", "--option-two"],
    )

    # Python was also verified
    create_command.verify_python.assert_called_once_with(first_app_config)

    # Reset the mock, then invoke verify_app_tools a second time.
    create_command.verify_python.reset_mock()
    create_command.verify_app_tools(app=first_app_config)

    # Python will *not* be verified a second time.
    create_command.verify_python.assert_not_called()



import pytest

from briefcase.exceptions import (
    BriefcaseCommandError,
    NetworkFailure,
    UnsupportedHostError,
)
from briefcase.integrations.linuxdeploy import LinuxDeployURLPlugin

from .utils import side_effect_create_mock_appimage


@pytest.mark.parametrize("host_os", ["Windows", "wonky"])
def test_unsupported_os(mock_tools, host_os):
    """When host OS is not supported, an error is raised."""
    mock_tools.host_os = host_os

    with pytest.raises(
        UnsupportedHostError,
        match=f"{LinuxDeployURLPlugin.name} is not supported on {host_os}",
    ):
        LinuxDeployURLPlugin.verify(mock_tools)


def test_verify(mock_tools, tmp_path):
    """URL plugins will be downloaded."""
    # Mock a successful download
    mock_tools.file.download.side_effect = side_effect_create_mock_appimage(
        tmp_path
        / "tools"
        / "linuxdeploy_plugins"
        / "sometool"
        / "f3355f8e631ffc1abbb7afd37b36315f7846182ca2276c481fb9a43a7f4d239f"
        / "linuxdeploy-plugin-sometool-wonky.AppImage"
    )

    LinuxDeployURLPlugin.verify(
        mock_tools,
        url="https://example.com/path/to/linuxdeploy-plugin-sometool-wonky.AppImage",
    )

    mock_tools.file.download.assert_called_with(
        url="https://example.com/path/to/linuxdeploy-plugin-sometool-wonky.AppImage",
        download_path=tmp_path
        / "tools"
        / "linuxdeploy_plugins"
        / "sometool"
        / "f3355f8e631ffc1abbb7afd37b36315f7846182ca2276c481fb9a43a7f4d239f",
        role="user-provided linuxdeploy plugin from URL",
    )


def test_download_failure(mock_tools, tmp_path):
    """A failure downloading a custom URL plugin raises an error."""

    # Mock a successful download
    mock_tools.file.download.side_effect = NetworkFailure("mock")

    url = "https://example.com/path/to/linuxdeploy-plugin-sometool-wonky.AppImage"

    with pytest.raises(NetworkFailure, match="Unable to mock"):
        LinuxDeployURLPlugin.verify(mock_tools, url=url)

    # A download was invoked
    mock_tools.file.download.assert_called_with(
        url=url,
        download_path=mock_tools.base_path
        / "linuxdeploy_plugins"
        / "sometool"
        / "f3355f8e631ffc1abbb7afd37b36315f7846182ca2276c481fb9a43a7f4d239f",
        role="user-provided linuxdeploy plugin from URL",
    )


def test_invalid_plugin_name(mock_tools, tmp_path):
    """If the URL filename doesn't match the pattern of a linuxdeploy plugin, an error
    is raised."""

    with pytest.raises(BriefcaseCommandError):
        LinuxDeployURLPlugin.verify(
            mock_tools,
            url="https://example.com/path/to/not-a-plugin.exe",
        )



import pytest

from briefcase.console import Console, Log
from briefcase.platforms.windows.app import WindowsAppCreateCommand


@pytest.fixture
def create_command(tmp_path):
    return WindowsAppCreateCommand(
        logger=Log(),
        console=Console(),
        base_path=tmp_path / "base_path",
        data_path=tmp_path / "briefcase",
    )


def test_binary_path(create_command, first_app_config, tmp_path):
    binary_path = create_command.binary_path(first_app_config)

    expected_path = (
        tmp_path
        / "base_path"
        / "build"
        / "first-app"
        / "windows"
        / "app"
        / "src"
        / "First App.exe"
    )
    assert binary_path == expected_path


def test_project_path(create_command, first_app_config, tmp_path):
    """The project path is the bundle path."""
    project_path = create_command.project_path(first_app_config)
    bundle_path = create_command.bundle_path(first_app_config)

    expected_path = tmp_path / "base_path/build/first-app/windows/app"
    assert expected_path == project_path == bundle_path


def test_distribution_path(create_command, first_app_config, tmp_path):
    distribution_path = create_command.distribution_path(first_app_config)

    expected_path = tmp_path / "base_path/dist/First App-0.0.1.msi"
    assert distribution_path == expected_path



import sys
from pathlib import Path, PurePosixPath

import pytest


@pytest.mark.skipif(
    sys.platform == "win32", reason="Windows paths aren't converted in Docker context"
)
@pytest.mark.parametrize(
    "original, path_map, dockerized",
    [
        # Simple values are unmodified
        ("value", None, "value"),
        # sys.executable is replaced with the specified version
        (
            "/path/to/python",
            {"/path/to/python": "python3.X", "/base/briefcase": "/briefcase"},
            "python3.X",
        ),
        # bundle path is replaced with /app
        (
            "/base/bundle/path/to/file",
            {"/base/bundle": "/app", "/base/briefcase": "/briefcase"},
            "/app/path/to/file",
        ),
        # data path is replaced with /briefcase
        (
            "/base/briefcase/path/to/file",
            {"/base/bundle": "/app", "/base/briefcase": "/briefcase"},
            "/briefcase/path/to/file",
        ),
        # Path inputs are converted
        (
            PurePosixPath("/base/briefcase/path/to/file"),
            {"/base/bundle": "/app", "/base/briefcase": "/briefcase"},
            "/briefcase/path/to/file",
        ),
        # Multiple references in a single path are converted
        (
            "/unmodified/path:/base/bundle/path/to/file:/base/briefcase/path/to/other/file",
            {"/base/bundle": "/app", "/base/briefcase": "/briefcase"},
            "/unmodified/path:/app/path/to/file:/briefcase/path/to/other/file",
        ),
        # path_map supports Path objects
        (
            "/unmodified/path:/base/bundle/path/to/file:/base/briefcase/path/to/other/file",
            {
                PurePosixPath("/base/bundle"): Path("/app"),
                Path("/base/briefcase"): PurePosixPath("/briefcase"),
            },
            "/unmodified/path:/app/path/to/file:/briefcase/path/to/other/file",
        ),
    ],
)
def test_dockerize_path(
    mock_docker,
    original,
    path_map,
    dockerized,
    tmp_path,
    monkeypatch,
):
    monkeypatch.setattr("sys.executable", "/path/to/python")

    converted_path = mock_docker.dockerize_path(
        original,
        path_map=path_map,
    )

    assert converted_path == dockerized



from __future__ import annotations

import plistlib
import subprocess
import time
from pathlib import Path
from uuid import UUID

from briefcase.commands import (
    BuildCommand,
    CreateCommand,
    OpenCommand,
    PackageCommand,
    PublishCommand,
    RunCommand,
    UpdateCommand,
)
from briefcase.config import AppConfig
from briefcase.console import InputDisabled, select_option
from briefcase.exceptions import (
    BriefcaseCommandError,
    InvalidDeviceError,
    NoDistributionArtefact,
)
from briefcase.integrations.subprocess import is_process_dead
from briefcase.integrations.xcode import DeviceState, get_device_state, get_simulators
from briefcase.platforms.iOS import iOSMixin
from briefcase.platforms.macOS.filters import XcodeBuildFilter, macOS_log_clean_filter


class iOSXcodePassiveMixin(iOSMixin):
    output_format = "Xcode"

    @property
    def packaging_formats(self):
        return ["ipa"]

    @property
    def default_packaging_format(self):
        return "ipa"

    def project_path(self, app):
        return self.bundle_path(app) / f"{app.formal_name}.xcodeproj"

    def binary_path(self, app):
        return (
            self.bundle_path(app)
            / "build"
            / "Debug-iphonesimulator"
            / f"{app.formal_name}.app"
        )

    def distribution_path(self, app):
        # This path won't ever be *generated*, as distribution artefacts
        # can't be generated on iOS.
        raise NoDistributionArtefact(
            """
*************************************************************************
** WARNING: No distributable artefact has been generated               **
*************************************************************************

    Briefcase has not generated a standalone iOS artefact, as iOS apps
    must be published through Xcode.

    To open Xcode for your iOS project, run:

        briefcase open iOS

    and use Xcode's app distribution workflow.

*************************************************************************
"""
        )


class iOSXcodeMixin(iOSXcodePassiveMixin):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # External service APIs.
        # These are abstracted to enable testing without patching.
        self.get_simulators = get_simulators

    def add_options(self, parser):
        super().add_options(parser)
        parser.add_argument(
            "-d",
            "--device",
            dest="udid",
            help="The device to target; either a UDID, "
            'a device name ("iPhone 11"), '
            'or a device name and OS version ("iPhone 11::iOS 13.3")',
            required=False,
        )

    def select_target_device(self, udid_or_device=None):
        """Select the target device to use for iOS builds.

        Interrogates the system to get the list of available simulators

        If there is only a single iOS version available, that version
        will be selected automatically.

        If there is only one simulator available, that version will be selected
        automatically.

        If the user has specified a device at the command line, it will be
        used in preference to any

        :param udid_or_device: The device to target. Can be a device UUID, a
            device name ("iPhone 11"), or a device name and OS version
            ("iPhone 11::13.3"). If ``None``, the user will be asked to select
            a device at runtime.
        :returns: A tuple containing the udid, iOS version, and device name
            for the selected device.
        """
        simulators = self.get_simulators(self.tools, "iOS")
        try:
            # Try to convert to a UDID. If this succeeds, then the argument
            # is a UDID.
            udid = str(UUID(udid_or_device)).upper()
            # User has provided a UDID at the command line; look for it.
            for iOS_tag, devices in simulators.items():
                try:
                    device = devices[udid]
                    # iOS_tag will be of the form "iOS 15.5"
                    # Drop the "iOS" prefix when reporting the version.
                    iOS_version = iOS_tag.split(" ", 1)[-1]
                    return udid, iOS_version, device
                except KeyError:
                    # UDID doesn't exist in this iOS version; try another.
                    pass

            # We've iterated through all available iOS versions and
            # found no match; return an error.
            raise InvalidDeviceError("device UDID", udid)

        except (ValueError, TypeError) as e:
            # Provided value wasn't a UDID.
            # It must be a device or device+version
            if udid_or_device and "::" in udid_or_device:
                # A device name::version.
                device, iOS_tag = udid_or_device.split("::")

                try:
                    # Convert the simulator dict into a dict where
                    # the iOS versions are lower cased, then do a lookup
                    # on the lower case name provided by the user.
                    # However, also return the *unmodified* iOS version string
                    # so we can convert the user-provided iOS version into the
                    # "clean", official capitalization.
                    iOS_tag, devices = {
                        clean_iOS_tag.lower(): (clean_iOS_tag, details)
                        for clean_iOS_tag, details in simulators.items()
                    }[iOS_tag.lower()]

                    # iOS_tag will be of the form "iOS 15.5"
                    # Drop the "iOS" prefix when reporting the version.
                    iOS_version = iOS_tag.split(" ", 1)[-1]
                    try:
                        # Do a reverse lookup for UDID, based on a
                        # case-insensitive name lookup.
                        udid = {name.lower(): udid for udid, name in devices.items()}[
                            device.lower()
                        ]

                        # Found a match;
                        # normalize back to the official name and return.
                        device = devices[udid]
                        return udid, iOS_version, device
                    except KeyError as e:
                        raise InvalidDeviceError("device name", device) from e
                except KeyError as e:
                    raise InvalidDeviceError("iOS Version", iOS_tag) from e
            elif udid_or_device:
                # Just a device name
                device = udid_or_device

                # Search iOS versions, looking for most recent version first.
                # The iOS tag will be something like "iOS 15.4";
                # Drop the prefix (if it exists), convert into the tuple (15, 4),
                # and sort numerically.
                for iOS_tag, devices in sorted(
                    simulators.items(),
                    key=lambda item: tuple(
                        int(v) for v in item[0].split()[-1].split(".")
                    ),
                    reverse=True,
                ):
                    try:
                        udid = {name.lower(): udid for udid, name in devices.items()}[
                            device.lower()
                        ]

                        # Found a match;
                        # normalize back to the official name and return.
                        device = devices[udid]

                        # iOS_tag will be of the form "iOS 15.5"
                        # Drop the "iOS" prefix when reporting the version.
                        iOS_version = iOS_tag.split(" ", 1)[-1]

                        return udid, iOS_version, device
                    except KeyError:
                        # UDID doesn't exist in this iOS version; try another.
                        pass
                raise InvalidDeviceError("device name", device) from e

        if len(simulators) == 0:
            raise BriefcaseCommandError("No iOS simulators available.")
        elif len(simulators) == 1:
            iOS_tag = list(simulators.keys())[0]
        else:
            self.input.prompt()
            self.input.prompt("Select iOS version:")
            self.input.prompt()
            iOS_tag = select_option(
                {tag: tag for tag in simulators.keys()}, input=self.input
            )

        devices = simulators[iOS_tag]

        if len(devices) == 0:
            raise BriefcaseCommandError(f"No simulators available for {iOS_tag}.")
        elif len(devices) == 1:
            udid = list(devices.keys())[0]
        else:
            self.input.prompt()
            self.input.prompt("Select simulator device:")
            self.input.prompt()
            udid = select_option(devices, input=self.input)

        device = devices[udid]

        self.logger.info(
            f"""
In the future, you could specify this device by running:

    $ briefcase {self.command} iOS -d "{device}::{iOS_tag}"

or:

    $ briefcase {self.command} iOS -d {udid}
"""
        )

        # iOS_tag will be of the form "iOS 15.5"
        # Drop the "iOS" prefix when reporting the version.
        iOS_version = iOS_tag.split(" ", 1)[-1]

        return udid, iOS_version, device


class iOSXcodeCreateCommand(iOSXcodePassiveMixin, CreateCommand):
    description = "Create and populate a iOS Xcode project."

    def permissions_context(self, app: AppConfig, x_permissions: dict[str, str]):
        """Additional template context for permissions.

        :param app: The config object for the app
        :param x_permissions: The dictionary of known cross-platform permission
            definitions.
        :returns: The template context describing permissions for the app.
        """
        # The collection of info.plist entries
        info = {}

        if x_permissions["camera"]:
            info["NSCameraUsageDescription"] = x_permissions["camera"]
        if x_permissions["microphone"]:
            info["NSMicrophoneUsageDescription"] = x_permissions["microphone"]

        if x_permissions["fine_location"]:
            info["NSLocationWhenInUseUsageDescription"] = x_permissions["fine_location"]
            info["NSLocationDefaultAccuracyReduced"] = False
        elif x_permissions["coarse_location"]:
            info["NSLocationWhenInUseUsageDescription"] = x_permissions[
                "coarse_location"
            ]
            info["NSLocationDefaultAccuracyReduced"] = True

        if x_permissions["background_location"]:
            if "NSLocationWhenInUseUsageDescription" not in info:
                info["NSLocationWhenInUseUsageDescription"] = x_permissions[
                    "background_location"
                ]
            info["NSLocationAlwaysAndWhenInUseUsageDescription"] = x_permissions[
                "background_location"
            ]
            info["UIBackgroundModes"] = ["processing", "location"]

        if x_permissions["photo_library"]:
            info["NSPhotoLibraryAddUsageDescription"] = x_permissions["photo_library"]

        # Override any info.plist entries with the platform specific definitions
        info.update(getattr(app, "info", {}))

        return {
            "info": info,
        }

    def _extra_pip_args(self, app: AppConfig):
        """Any additional arguments that must be passed to pip when installing packages.

        :param app: The app configuration
        :returns: A list of additional arguments
        """
        return [
            "--prefer-binary",
            "--extra-index-url",
            "https://pypi.anaconda.org/beeware/simple",
        ]

    def _install_app_requirements(
        self,
        app: AppConfig,
        requires: list[str],
        app_packages_path: Path,
    ):
        # Perform the initial install pass targeting the "iphoneos" platform
        super()._install_app_requirements(
            app,
            requires=requires,
            app_packages_path=app_packages_path.parent / "app_packages.iphoneos",
            progress_message="Installing app requirements for iPhone device...",
            pip_kwargs={
                "env": {
                    "PYTHONPATH": str(
                        self.support_path(app) / "platform-site/iphoneos.arm64"
                    ),
                }
            },
        )

        # Perform a second install pass targeting the "iphonesimulator" platform for the
        # current architecture
        super()._install_app_requirements(
            app,
            requires=requires,
            app_packages_path=app_packages_path.parent / "app_packages.iphonesimulator",
            progress_message="Installing app requirements for iPhone simulator...",
            pip_kwargs={
                "env": {
                    "PYTHONPATH": str(
                        self.support_path(app)
                        / "platform-site"
                        / f"iphonesimulator.{self.tools.host_arch}"
                    ),
                }
            },
        )


class iOSXcodeUpdateCommand(iOSXcodeCreateCommand, UpdateCommand):
    description = "Update an existing iOS Xcode project."


class iOSXcodeOpenCommand(iOSXcodePassiveMixin, OpenCommand):
    description = "Open an existing iOS Xcode project."


class iOSXcodeBuildCommand(iOSXcodePassiveMixin, BuildCommand):
    description = "Build an iOS Xcode project."

    def info_plist_path(self, app: AppConfig):
        """Obtain the path to the application's plist file.

        :param app: The config object for the app
        :return: The full path of the application's plist file.
        """
        return self.bundle_path(app) / self.path_index(app, "info_plist_path")

    def update_app_metadata(self, app: AppConfig, test_mode: bool):
        with self.input.wait_bar("Setting main module..."):
            # Load the original plist
            with self.info_plist_path(app).open("rb") as f:
                info_plist = plistlib.load(f)

            # Set the name of the app's main module; this will depend
            # on whether we're in test mode.
            info_plist["MainModule"] = app.main_module(test_mode)

            # Write the modified plist
            with self.info_plist_path(app).open("wb") as f:
                plistlib.dump(info_plist, f)

    def build_app(self, app: AppConfig, test_mode: bool = False, **kwargs):
        """Build the Xcode project for the application.

        :param app: The application to build
        :param test_mode: Should the app be updated in test mode? (default: False)
        """
        self.logger.info("Updating app metadata...", prefix=app.app_name)
        self.update_app_metadata(app=app, test_mode=test_mode)

        self.logger.info("Building Xcode project...", prefix=app.app_name)
        with self.input.wait_bar("Building..."):
            try:
                self.tools.subprocess.run(
                    [
                        "xcodebuild",
                        "build",
                        "-project",
                        self.project_path(app),
                        "-destination",
                        'platform="iOS Simulator"',
                        "-configuration",
                        "Debug",
                        "-arch",
                        self.tools.host_arch,
                        "-sdk",
                        "iphonesimulator",
                        "-verbose" if self.tools.logger.is_deep_debug else "-quiet",
                    ],
                    check=True,
                    filter_func=(
                        None if self.tools.logger.is_deep_debug else XcodeBuildFilter()
                    ),
                )
            except subprocess.CalledProcessError as e:
                raise BriefcaseCommandError(
                    f"Unable to build app {app.app_name}."
                ) from e


class iOSXcodeRunCommand(iOSXcodeMixin, RunCommand):
    description = "Run an iOS Xcode project."

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # External service APIs.
        # This is abstracted to enable testing without patching.
        self.get_device_state = get_device_state

    def run_app(
        self,
        app: AppConfig,
        test_mode: bool,
        passthrough: list[str],
        udid=None,
        **kwargs,
    ):
        """Start the application.

        :param app: The config object for the app
        :param test_mode: Boolean; Is the app running in test mode?
        :param passthrough: The list of arguments to pass to the app
        :param udid: The device UDID to target. If ``None``, the user will
            be asked to select a device at runtime.
        """
        try:
            udid, iOS_version, device = self.select_target_device(udid)
        except InputDisabled as e:
            raise BriefcaseCommandError(
                "Input has been disabled; can't select a device to target."
            ) from e

        if test_mode:
            label = "test suite"
        else:
            label = "app"

        self.logger.info(
            f"Starting {label} on an {device} running iOS {iOS_version} (device UDID {udid})",
            prefix=app.app_name,
        )

        # The simulator needs to be booted before being started.
        # If it's shut down, we can boot it again; but if it's currently
        # shutting down, we need to wait for it to shut down before restarting.
        device_state = self.get_device_state(self.tools, udid)
        if device_state not in {DeviceState.SHUTDOWN, DeviceState.BOOTED}:
            with self.input.wait_bar("Waiting for simulator shutdown..."):
                while device_state not in {DeviceState.SHUTDOWN, DeviceState.BOOTED}:
                    time.sleep(2)
                    device_state = self.get_device_state(self.tools, udid)

        # We now know the simulator is either shut down or booted;
        # if it's shut down, start it again.
        if device_state == DeviceState.SHUTDOWN:
            try:
                with self.input.wait_bar("Booting simulator..."):
                    self.tools.subprocess.run(
                        ["xcrun", "simctl", "boot", udid],
                        check=True,
                    )
            except subprocess.CalledProcessError as e:
                raise BriefcaseCommandError(
                    f"Unable to boot {device} simulator running {iOS_version}"
                ) from e

        if not test_mode:
            # We now know the simulator is *running*, so we can open it.
            # We don't need to open the simulator to run the test suite.
            try:
                with self.input.wait_bar("Opening simulator..."):
                    self.tools.subprocess.run(
                        [
                            "open",
                            "-a",
                            "Simulator",
                            "--args",
                            "-CurrentDeviceUDID",
                            udid,
                        ],
                        check=True,
                    )
            except subprocess.CalledProcessError as e:
                raise BriefcaseCommandError(
                    f"Unable to open {device} simulator running {iOS_version}"
                ) from e

        # Try to uninstall the app first. If the app hasn't been installed
        # before, this will still succeed.
        self.logger.info(f"Installing {label}...", prefix=app.app_name)
        with self.input.wait_bar(
            "Uninstalling any existing app version..."
        ) as keep_alive, self.tools.subprocess.Popen(
            ["xcrun", "simctl", "uninstall", udid, app.bundle_identifier]
        ) as uninstall_popen:
            while (ret_code := uninstall_popen.poll()) is None:
                keep_alive.update()
                time.sleep(0.25)
        if ret_code != 0:
            self.logger.error(f"{ret_code=}")
            raise BriefcaseCommandError(
                f"Unable to uninstall old version of app {app.app_name}."
            )

        # Install the app.
        with self.input.wait_bar(
            f"Installing new {label} version..."
        ) as keep_alive, self.tools.subprocess.Popen(
            ["xcrun", "simctl", "install", udid, self.binary_path(app)]
        ) as install_popen:
            while (ret_code := install_popen.poll()) is None:
                keep_alive.update()
                time.sleep(0.25)
        if ret_code != 0:
            raise BriefcaseCommandError(
                f"Unable to install new version of app {app.app_name}."
            )

        # Start log stream for the app.
        # The following sets up a log stream filter that looks for:
        #  1. a log sender that matches that app binary; or,
        #  2. a log sender of that is a Python extension module,
        #     and a process that matches the app binary.
        # Case (1) works when the standard library is statically linked,
        #   and for native NSLog() calls in the bootstrap binary
        # Case (2) works when the standard library is dynamically linked,
        #   and ctypes (which handles the NSLog integration) is an
        #   extension module.
        # It's not enough to filter on *just* the processImagePath,
        # as the process will generate lots of system-level messages.
        # We can't filter on *just* the senderImagePath, because other
        # apps will generate log messages that would be caught by the filter.
        simulator_log_popen = self.tools.subprocess.Popen(
            [
                "xcrun",
                "simctl",
                "spawn",
                udid,
                "log",
                "stream",
                "--style",
                "compact",
                "--predicate",
                f'senderImagePath ENDSWITH "/{app.formal_name}"'
                f' OR (processImagePath ENDSWITH "/{app.formal_name}"'
                ' AND (senderImagePath ENDSWITH "-iphonesimulator.so"'
                ' OR senderImagePath ENDSWITH "-iphonesimulator.dylib"))',
            ],
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            bufsize=1,
        )

        # Wait for the log stream start up
        time.sleep(0.25)

        try:
            self.logger.info(f"Starting {label}...", prefix=app.app_name)
            with self.input.wait_bar(f"Launching {label}..."):
                output = self.tools.subprocess.check_output(
                    ["xcrun", "simctl", "launch", udid, app.bundle_identifier]
                    + passthrough
                )
                try:
                    app_pid = int(output.split(":")[1].strip())
                except (IndexError, ValueError) as e:
                    raise BriefcaseCommandError(
                        f"Unable to determine PID of {label} {app.app_name}."
                    ) from e

            # Start streaming logs for the app.
            self.logger.info(
                "Following simulator log output (type CTRL-C to stop log)...",
                prefix=app.app_name,
            )

            # Stream the app logs,
            self._stream_app_logs(
                app,
                popen=simulator_log_popen,
                test_mode=test_mode,
                clean_filter=macOS_log_clean_filter,
                clean_output=True,
                stop_func=lambda: is_process_dead(app_pid),
                log_stream=True,
            )
        except subprocess.CalledProcessError as e:
            raise BriefcaseCommandError(
                f"Unable to launch {label} {app.app_name}."
            ) from e

        # Preserve the device selection as state.
        return {"udid": udid}


class iOSXcodePackageCommand(iOSXcodeMixin, PackageCommand):
    description = "Package an iOS app."


class iOSXcodePublishCommand(iOSXcodeMixin, PublishCommand):
    description = "Publish an iOS app."
    publication_channels = ["ios_appstore"]
    default_publication_channel = "ios_appstore"


# Declare the briefcase command bindings
create = iOSXcodeCreateCommand
update = iOSXcodeUpdateCommand
open = iOSXcodeOpenCommand
build = iOSXcodeBuildCommand
run = iOSXcodeRunCommand
package = iOSXcodePackageCommand
publish = iOSXcodePublishCommand



from __future__ import annotations

from unittest import mock

import pytest
import tomli_w
from cookiecutter.main import cookiecutter

from briefcase.commands import CreateCommand
from briefcase.config import AppConfig
from briefcase.console import Console, Log
from briefcase.integrations.base import Tool
from briefcase.integrations.subprocess import Subprocess

from ...utils import DummyConsole, create_file


@pytest.fixture
def monkeypatch_tool_host_os(monkeypatch):
    """Add testing host OS as supported for tools that support all platforms."""
    monkeypatch.setattr(
        Tool,
        "supported_host_os",
        Tool.supported_host_os.union({"c64"}),
    )


class DefaultCreateCommand(CreateCommand):
    # An instance of CreateCommand that inherits the default
    # behavior of create handling.

    # method is required by the interface, but are not needed for these tests.
    def binary_path(self, app):
        return NotImplementedError()


@pytest.fixture
def default_create_command(tmp_path):
    return DefaultCreateCommand(base_path=tmp_path, logger=Log(), console=Console())


class DummyCreateCommand(CreateCommand):
    """A dummy create command that stubs out all the required interfaces of the Create
    command."""

    supported_host_os = {"c64"}
    # Platform and format contain upper case to test case normalization
    platform = "Tester"
    output_format = "Dummy"
    description = "Dummy create command"
    hidden_app_properties = {"permission", "request"}

    def __init__(self, *args, support_file=None, git=None, home_path=None, **kwargs):
        kwargs.setdefault("logger", Log())
        kwargs.setdefault("console", Console())
        super().__init__(*args, **kwargs)

        # Override the host properties
        self.tools.host_arch = "gothic"
        self.tools.host_os = "c64"

        self.tools.home_path = home_path

        # If a test sets this property, the tool verification step will fail.
        self._missing_tool = None

        # Mock the external services
        self.tools.git = git
        self.tools.subprocess = mock.MagicMock(spec_set=Subprocess)
        self.support_file = support_file
        self.tools.input = DummyConsole()
        self.tools.cookiecutter = mock.MagicMock(spec_set=cookiecutter)

    @property
    def support_package_url_query(self):
        """The query arguments to use in a support package query request."""
        return [
            ("platform", self.platform),
            ("version", self.python_version_tag),
            ("arch", self.tools.host_arch),
        ]

    def binary_path(self, app):
        return self.bundle_path(app) / f"{app.app_name}.bin"

    # Hard code the python version to make testing easier.
    @property
    def python_version_tag(self):
        return "3.X"

    # Define output format-specific template context.
    def output_format_template_context(self, app):
        return {"output_format": "dummy"}

    # Handle platform-specific permissions.
    # Convert all the cross-platform permissions to upper case, prefixing DUMMY_.
    # Add a "good lighting" request if the camera permission has been requested.
    def permissions_context(self, app: AppConfig, x_permissions: dict[str, str]):
        # We don't actually need anything from the superclass; but call it to ensure
        # coverage.
        context = super().permissions_context(app, x_permissions)
        if context:
            # Make sure the base class *isn't* doing anything.
            return context

        permissions = {
            f"DUMMY_{key.upper()}": value.upper()
            for key, value in x_permissions.items()
            if value
        }
        context["permissions"] = permissions
        context["custom_permissions"] = app.permission

        requests = {"good.lighting": True} if x_permissions["camera"] else {}
        requests.update(getattr(app, "request", {}))
        context["requests"] = requests

        return context


class TrackingCreateCommand(DummyCreateCommand):
    """A dummy creation command that doesn't actually do anything.

    It only serves to track which actions would be performed.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.actions = []

    def briefcase_toml(self, app):
        # default any app to an empty `briefcase.toml`
        return self._briefcase_toml.get(app, {})

    def verify_host(self):
        super().verify_host()
        self.actions.append(("verify-host",))

    def verify_tools(self):
        super().verify_tools()
        self.actions.append(("verify-tools",))

    def finalize_app_config(self, app):
        super().finalize_app_config(app=app)
        self.actions.append(("finalize-app-config", app.app_name))

    def verify_app_template(self, app):
        super().verify_app_template(app=app)
        self.actions.append(("verify-app-template", app.app_name))

    def verify_app_tools(self, app):
        super().verify_app_tools(app=app)
        self.actions.append(("verify-app-tools", app.app_name))

    # Override all the body methods of a CreateCommand
    # with versions that we can use to track actions performed.
    def generate_app_template(self, app):
        self.actions.append(("generate", app.app_name))

        # A mock version of template generation.
        create_file(self.bundle_path(app) / "new", "new template!")

    def install_app_support_package(self, app):
        self.actions.append(("support", app.app_name))

    def install_app_requirements(self, app, test_mode):
        self.actions.append(("requirements", app.app_name, test_mode))

    def install_app_code(self, app, test_mode):
        self.actions.append(("code", app.app_name, test_mode))

    def install_app_resources(self, app):
        self.actions.append(("resources", app.app_name))

    def install_stub_binary(self, app):
        self.actions.append(("stub", app.app_name))
        # A mock version of a stub binary
        create_file(self.bundle_path(app) / "Stub.bin", "stub binary")

    def cleanup_app_content(self, app):
        self.actions.append(("cleanup", app.app_name))


@pytest.fixture
def create_command(tmp_path, mock_git, monkeypatch_tool_host_os):
    return DummyCreateCommand(
        base_path=tmp_path / "base_path",
        data_path=tmp_path / "data",
        git=mock_git,
        home_path=tmp_path / "home",
    )


@pytest.fixture
def tracking_create_command(tmp_path, mock_git, monkeypatch_tool_host_os):
    return TrackingCreateCommand(
        git=mock_git,
        base_path=tmp_path / "base_path",
        apps={
            "first": AppConfig(
                app_name="first",
                bundle="com.example",
                version="0.0.1",
                description="The first simple app",
                sources=["src/first"],
                license={"file": "LICENSE"},
            ),
            "second": AppConfig(
                app_name="second",
                bundle="com.example",
                version="0.0.2",
                description="The second simple app",
                sources=["src/second"],
                license={"file": "LICENSE"},
            ),
        },
    )


@pytest.fixture
def myapp():
    return AppConfig(
        app_name="my-app",
        formal_name="My App",
        bundle="com.example",
        version="1.2.3",
        description="This is a simple app",
        sources=["src/my_app"],
        url="https://example.com",
        author="First Last",
        author_email="first@example.com",
        license={"file": "LICENSE"},
    )


@pytest.fixture
def bundle_path(myapp, tmp_path):
    # Return the bundle path for the app; however, as a side effect,
    # ensure that the app, and app_packages target directories
    # exist, and the briefcase index file has been created.
    bundle_path = tmp_path / "base_path/build" / myapp.app_name / "tester/dummy"
    (bundle_path / "path/to/app").mkdir(parents=True, exist_ok=True)

    return bundle_path


@pytest.fixture
def app_packages_path_index(bundle_path):
    (bundle_path / "path/to/app_packages").mkdir(parents=True, exist_ok=True)
    with (bundle_path / "briefcase.toml").open("wb") as f:
        index = {
            "paths": {
                "app_path": "path/to/app",
                "app_packages_path": "path/to/app_packages",
                "support_path": "path/to/support",
                "support_revision": 37,
            }
        }
        tomli_w.dump(index, f)


@pytest.fixture
def app_requirements_path_index(bundle_path):
    with (bundle_path / "briefcase.toml").open("wb") as f:
        index = {
            "paths": {
                "app_path": "path/to/app",
                "app_requirements_path": "path/to/requirements.txt",
                "support_path": "path/to/support",
                "support_revision": 37,
            }
        }
        tomli_w.dump(index, f)


@pytest.fixture
def no_support_revision_index(bundle_path):
    with (bundle_path / "briefcase.toml").open("wb") as f:
        index = {
            "paths": {
                "app_path": "path/to/app",
                "app_requirements_path": "path/to/requirements.txt",
                "support_path": "path/to/support",
            }
        }
        tomli_w.dump(index, f)


@pytest.fixture
def no_support_path_index(bundle_path):
    with (bundle_path / "briefcase.toml").open("wb") as f:
        index = {
            "paths": {
                "app_path": "path/to/app",
                "app_requirements_path": "path/to/requirements.txt",
            }
        }
        tomli_w.dump(index, f)


@pytest.fixture
def support_path(bundle_path):
    return bundle_path / "path/to/support"


@pytest.fixture
def app_requirements_path(bundle_path):
    return bundle_path / "path/to/requirements.txt"


@pytest.fixture
def app_packages_path(bundle_path):
    return bundle_path / "path/to/app_packages"


@pytest.fixture
def app_path(bundle_path):
    return bundle_path / "path/to/app"


@pytest.fixture
def stub_binary_revision_path_index(bundle_path):
    with (bundle_path / "briefcase.toml").open("wb") as f:
        index = {
            "paths": {
                "app_path": "path/to/app",
                "app_requirements_path": "path/to/requirements.txt",
                "stub_binary_revision": 37,
            }
        }
        tomli_w.dump(index, f)



import pytest

from briefcase.exceptions import BriefcaseCommandError


def create_file_and_parents(path, content):
    path.parent.mkdir(parents=True, exist_ok=True)
    path.write_text(content, encoding="utf-8")


def test_src_is_used_by_default(convert_command):
    module_path = convert_command.base_path / "src" / "module_name"
    create_file_and_parents(module_path / "__main__.py", "")
    default, info = convert_command.get_source_dir_hint("module-name", "module_name")
    assert default == str(module_path.relative_to(convert_command.base_path)).replace(
        "\\", "/"
    )
    assert repr(default) in info


def test_flat_is_used_if_src_name_is_wrong(convert_command):
    module_path = convert_command.base_path / "module_name"
    create_file_and_parents(module_path / "__main__.py", "")
    default, info = convert_command.get_source_dir_hint("module-name", "module_name")
    assert default == str(module_path.relative_to(convert_command.base_path)).replace(
        "\\", "/"
    )
    assert repr(default) in info


def test_fallback_is_used_if_no_src_or_flat(convert_command):
    module_path = convert_command.base_path / "something/module_name"
    create_file_and_parents(module_path / "__main__.py", "")
    default, info = convert_command.get_source_dir_hint("module-name", "module_name")
    assert default == str(module_path.relative_to(convert_command.base_path)).replace(
        "\\", "/"
    )
    assert repr(default) in info


def test_src_is_preferred_over_flat(convert_command):
    module_path = convert_command.base_path / "src" / "module_name"
    create_file_and_parents(module_path / "__main__.py", "")
    flat_module_path = convert_command.base_path / "module_name"
    create_file_and_parents(module_path / "__main__.py", "")
    default, info = convert_command.get_source_dir_hint("module-name", "module_name")
    assert default == str(module_path.relative_to(convert_command.base_path)).replace(
        "\\", "/"
    )
    assert repr(default) in info
    assert repr(flat_module_path) not in info


def test_flat_is_preferred_over_fallback(convert_command):
    module_path = convert_command.base_path / "module_name"
    create_file_and_parents(module_path / "__main__.py", "")
    fallback_module_path = convert_command.base_path / "something/module_name"
    create_file_and_parents(module_path / "__main__.py", "")
    default, info = convert_command.get_source_dir_hint("module-name", "module_name")
    assert default == str(module_path.relative_to(convert_command.base_path)).replace(
        "\\", "/"
    )
    assert repr(default) in info
    assert repr(fallback_module_path) not in info


def test_flat_is_used_if_src_is_wrong(convert_command):
    module_path = convert_command.base_path / "module_name"
    create_file_and_parents(module_path / "__main__.py", "")
    wrong_module_path = convert_command.base_path / "src" / "different_module_name"
    create_file_and_parents(wrong_module_path / "__main__.py", "")
    default, info = convert_command.get_source_dir_hint("module-name", "module_name")
    assert default == str(module_path.relative_to(convert_command.base_path)).replace(
        "\\", "/"
    )
    assert repr(default) in info
    assert repr(module_path.name) in info


def test_flat_is_used_if_src_doesnt_have_main_file(convert_command):
    module_path = convert_command.base_path / "module_name"
    create_file_and_parents(module_path / "__main__.py", "")
    wrong_module_path = convert_command.base_path / "src" / "module_name"
    create_file_and_parents(wrong_module_path / "not_main.py", "")
    default, info = convert_command.get_source_dir_hint("module-name", "module_name")
    assert default == str(module_path.relative_to(convert_command.base_path)).replace(
        "\\", "/"
    )
    assert repr(default) in info
    assert repr(module_path.name) in info


def test_fallback_is_used_if_flat_doesnt_have_main_file(convert_command):
    module_path = convert_command.base_path / "something" / "module_name"
    create_file_and_parents(module_path / "__main__.py", "")
    wrong_module_path = convert_command.base_path / "module_name"
    create_file_and_parents(wrong_module_path / "not_main.py", "")
    default, info = convert_command.get_source_dir_hint("module-name", "module_name")
    assert default == str(module_path.relative_to(convert_command.base_path)).replace(
        "\\", "/"
    )
    assert repr(default) in info
    assert (
        repr(str(module_path.relative_to(convert_command.base_path))).replace(
            r"\\", "/"
        )
        in info
    )


def test_exception_is_raised_if_no_source_dir(convert_command):
    module_path = convert_command.base_path / "src" / "different_module_name"
    create_file_and_parents(module_path / "__main__.py", "")
    with pytest.raises(BriefcaseCommandError):
        convert_command.get_source_dir_hint("module-name", "module_name")


def test_exception_is_raised_if_no_source_dir_with_main(convert_command):
    module_path = convert_command.base_path / "src" / "module_name"
    create_file_and_parents(module_path / "not_main.py", "")
    with pytest.raises(BriefcaseCommandError):
        convert_command.get_source_dir_hint("module-name", "module_name")



import os
import shutil
import subprocess
from pathlib import Path
from unittest.mock import MagicMock, call

import pytest

from briefcase.exceptions import (
    BriefcaseCommandError,
    IncompatibleToolError,
    MissingToolError,
    NetworkFailure,
    UnsupportedHostError,
)
from briefcase.integrations.java import JDK

from ...utils import assert_url_resolvable, create_zip_file
from .conftest import JDK_BUILD, JDK_RELEASE

CALL_JAVA_HOME = call(["/usr/libexec/java_home"])


def test_short_circuit(mock_tools):
    """Tool is not created if already cached."""
    mock_tools.java = "tool"

    tool = JDK.verify(mock_tools)

    assert tool == "tool"
    assert tool == mock_tools.java


def test_unsupported_os(mock_tools):
    """When host OS is not supported, an error is raised."""
    mock_tools.host_os = "wonky"

    with pytest.raises(
        UnsupportedHostError,
        match=f"{JDK.name} is not supported on wonky",
    ):
        JDK.verify(mock_tools)


def test_unsupported_arch(mock_tools):
    """When the architecture is not supported, an error is raised."""
    mock_tools.host_arch = "IA-64"

    with pytest.raises(
        IncompatibleToolError,
        match="Briefcase cannot install Java JDK on this machine.",
    ):
        JDK.verify(mock_tools)


def test_macos_tool_java_home(mock_tools, capsys):
    """On macOS, the /usr/libexec/java_home utility is checked."""
    # Mock being on macOS
    mock_tools.host_os = "Darwin"

    # Mock 2 calls to check_output.
    mock_tools.subprocess.check_output.side_effect = [
        "/path/to/java",
        f"javac {JDK_RELEASE}\n",
    ]

    # Create a JDK wrapper by verification
    JDK.verify(mock_tools)

    # The JDK should have the path returned by the tool
    assert mock_tools.java.java_home == Path("/path/to/java")

    assert mock_tools.subprocess.check_output.mock_calls == [
        CALL_JAVA_HOME,
        # Second is a call to verify a valid Java version
        call([Path("/path/to/java/bin/javac"), "-version"]),
    ]

    # No console output
    output = capsys.readouterr()
    assert output.out == ""
    assert output.err == ""


def test_macos_tool_failure(mock_tools, tmp_path, capsys):
    """On macOS, if the libexec tool fails, the Briefcase JDK is used."""
    # Mock being on macOS
    mock_tools.host_os = "Darwin"

    # Mock a failed call on the libexec tool
    mock_tools.subprocess.check_output.side_effect = subprocess.CalledProcessError(
        returncode=1, cmd="/usr/libexec/java_home"
    )

    # Create a directory to make it look like the Briefcase Java already exists.
    (tmp_path / "tools/java17/Contents/Home/bin").mkdir(parents=True)

    # Create a JDK wrapper by verification
    jdk = JDK.verify(mock_tools)

    # The JDK should have the briefcase JAVA_HOME
    assert jdk.java_home == tmp_path / "tools/java17/Contents/Home"

    assert mock_tools.subprocess.check_output.mock_calls == [CALL_JAVA_HOME]

    # No console output
    output = capsys.readouterr()
    assert output.out == ""
    assert output.err == ""


def test_macos_wrong_jdk_version(mock_tools, tmp_path, capsys):
    """On macOS, if the libexec tool returns an unqualified JDK, the Briefcase JDK is
    used."""
    # Mock being on macOS
    mock_tools.host_os = "Darwin"

    # Mock 2 calls to check_output.
    mock_tools.subprocess.check_output.side_effect = [
        "/path/to/java",
        "java 1.8.0_352\n",
    ]

    # Create a directory to make it look like the Briefcase Java already exists.
    (tmp_path / "tools/java17/Contents/Home/bin").mkdir(parents=True)

    # Create a JDK wrapper by verification
    jdk = JDK.verify(mock_tools)

    # The JDK should have the briefcase JAVA_HOME
    assert jdk.java_home == tmp_path / "tools/java17/Contents/Home"

    assert mock_tools.subprocess.check_output.mock_calls == [
        CALL_JAVA_HOME,
        call([Path("/path/to/java/bin/javac"), "-version"]),
    ]

    # No console output
    output = capsys.readouterr()
    assert output.out == ""
    assert output.err == ""


def test_macos_invalid_jdk_path(mock_tools, tmp_path, capsys):
    """On macOS, if the libexec tool points to non-JDK path, the Briefcase JDK is
    used."""
    # Mock being on macOS
    mock_tools.host_os = "Darwin"

    # Mock 2 calls to check_output.
    mock_tools.subprocess.check_output.side_effect = [
        "/path/to/java",
        IndexError,
    ]

    # Create a directory to make it look like the Briefcase Java already exists.
    (tmp_path / "tools/java17/Contents/Home/bin").mkdir(parents=True)

    # Create a JDK wrapper by verification
    jdk = JDK.verify(mock_tools)

    # The JDK should have the briefcase JAVA_HOME
    assert jdk.java_home == tmp_path / "tools/java17/Contents/Home"

    assert mock_tools.subprocess.check_output.mock_calls == [
        CALL_JAVA_HOME,
        call([Path("/path/to/java/bin/javac"), "-version"]),
    ]

    # No console output
    output = capsys.readouterr()
    assert output.out == ""
    assert output.err == ""


def test_macos_provided_overrides_tool_java_home(mock_tools, capsys):
    """On macOS, an explicit JAVA_HOME overrides /usr/libexec/java_home."""
    # Mock being on macOS
    mock_tools.host_os = "Darwin"

    # Setup explicit JAVA_HOME
    mock_tools.os.environ = {"JAVA_HOME": "/path/to/java"}

    # Mock return value from javac. libexec won't be invoked.
    mock_tools.subprocess.check_output.return_value = f"javac {JDK_RELEASE}\n"

    # Create a JDK wrapper by verification
    JDK.verify(mock_tools)

    # The JDK should have the path returned by the tool
    assert mock_tools.java.java_home == Path("/path/to/java")

    # A single call to check output
    mock_tools.subprocess.check_output.assert_called_once_with(
        [Path("/path/to/java/bin/javac"), "-version"],
    ),

    # No console output
    output = capsys.readouterr()
    assert output.out == ""
    assert output.err == ""


def test_valid_provided_java_home(mock_tools, capsys):
    """If a valid JAVA_HOME is provided, it is used."""

    # Setup explicit JAVA_HOME
    mock_tools.os.environ = {"JAVA_HOME": "/path/to/java"}

    # Mock return value from javac.
    mock_tools.subprocess.check_output.return_value = f"javac {JDK_RELEASE}\n"

    # Create a JDK wrapper by verification
    JDK.verify(mock_tools)

    # The JDK should have the path returned by the tool
    assert mock_tools.java.java_home == Path("/path/to/java")

    # A single call to check output
    mock_tools.subprocess.check_output.assert_called_once_with(
        [Path("/path/to/java/bin/javac"), "-version"],
    ),

    # No console output
    output = capsys.readouterr()
    assert output.out == ""
    assert output.err == ""


@pytest.mark.parametrize(
    "host_os, java_home",
    [
        ("Linux", Path("tools", "java17")),
        ("Windows", Path("tools", "java17")),
        ("Darwin", Path("tools", "java17", "Contents", "Home")),
    ],
)
def test_invalid_jdk_version(mock_tools, host_os, java_home, tmp_path, capsys):
    """If the JDK pointed to by JAVA_HOME isn't a Java 8 JDK, the briefcase JDK is
    used."""
    # Mock os
    mock_tools.host_os = host_os

    # Setup explicit JAVA_HOME
    mock_tools.os.environ = {"JAVA_HOME": "/path/to/java"}

    # Mock return value from javac.
    mock_tools.subprocess.check_output.return_value = "javac 14\n"

    # Create a directory to make it look like the Briefcase Java already exists.
    (tmp_path / java_home / "bin").mkdir(parents=True)

    # Create a JDK wrapper by verification
    JDK.verify(mock_tools)

    # The JDK should have the briefcase JAVA_HOME
    assert mock_tools.java.java_home == tmp_path / java_home

    # A single call was made to check javac
    mock_tools.subprocess.check_output.assert_called_once_with(
        [Path("/path/to/java/bin/javac"), "-version"],
    )

    # Warning is shown for invalid JAVA_HOME
    output = capsys.readouterr()
    assert "WARNING: JAVA_HOME does not point to a Java 17 JDK" in output.out
    assert output.out.endswith("****\n")
    assert output.err == ""


@pytest.mark.parametrize(
    "host_os, java_home, error_type",
    [
        ("Linux", Path("tools", "java17"), FileNotFoundError),
        ("Linux", Path("tools", "java17"), NotADirectoryError),
        ("Windows", Path("tools", "java17"), FileNotFoundError),
        ("Windows", Path("tools", "java17"), NotADirectoryError),
        ("Darwin", Path("tools", "java17", "Contents", "Home"), FileNotFoundError),
        ("Darwin", Path("tools", "java17", "Contents", "Home"), NotADirectoryError),
    ],
)
def test_no_javac(mock_tools, host_os, java_home, error_type, tmp_path, capsys):
    """If the JAVA_HOME doesn't point to a location with a bin/javac, the briefcase JDK
    is used."""
    # Mock os
    mock_tools.host_os = host_os

    # Setup explicit JAVA_HOME
    mock_tools.os.environ = {"JAVA_HOME": "/path/to/nowhere"}

    # Mock return value from javac failing because executable doesn't exist
    # FileNotFoundError is raised if bin/javac doesn't exist.
    # NotADirectoryError is raised if the user-provided path in JAVA_HOME
    #   contains parts that exist in the filesystem but are not a directory.
    mock_tools.subprocess.check_output.side_effect = error_type

    # Create a directory to make it look like the Briefcase Java already exists.
    (tmp_path / java_home / "bin").mkdir(parents=True)

    # Create a JDK wrapper by verification
    JDK.verify(mock_tools)

    # The JAVA_HOME should point at the Briefcase-provided JDK
    assert mock_tools.java.java_home == tmp_path / java_home

    # A single call was made to check javac
    mock_tools.subprocess.check_output.assert_called_once_with(
        [Path("/path/to/nowhere/bin/javac"), "-version"],
    ),

    # Warning is shown for invalid JAVA_HOME
    output = capsys.readouterr()
    assert "WARNING: JAVA_HOME does not point to a JDK" in output.out
    assert output.out.endswith("****\n")
    assert output.err == ""


@pytest.mark.parametrize(
    "host_os, java_home",
    [
        ("Linux", Path("tools", "java17")),
        ("Windows", Path("tools", "java17")),
        ("Darwin", Path("tools", "java17", "Contents", "Home")),
    ],
)
def test_javac_error(mock_tools, host_os, java_home, tmp_path, capsys):
    """If javac can't be executed, the briefcase JDK is used."""
    # Mock os
    mock_tools.host_os = host_os

    # Setup explicit JAVA_HOME
    mock_tools.os.environ = {"JAVA_HOME": "/path/to/nowhere"}

    # Mock return value from javac failing because executable doesn't exist
    mock_tools.subprocess.check_output.side_effect = subprocess.CalledProcessError(
        returncode=1, cmd="/path/to/java/bin/javac"
    )

    # Create a directory to make it look like the Briefcase Java already exists.
    (tmp_path / java_home / "bin").mkdir(parents=True)

    # Create a JDK wrapper by verification
    JDK.verify(mock_tools)

    # The JDK should have the briefcase JAVA_HOME
    assert mock_tools.java.java_home == tmp_path / java_home

    # A single call was made to check javac
    mock_tools.subprocess.check_output.assert_called_once_with(
        [Path("/path/to/nowhere/bin/javac"), "-version"],
    ),

    # Warning is shown for invalid JAVA_HOME
    output = capsys.readouterr()
    assert "WARNING: Unable to invoke the Java compiler" in output.out
    assert output.out.endswith("****\n")
    assert output.err == ""


@pytest.mark.parametrize(
    "host_os, java_home",
    [
        ("Linux", Path("tools", "java17")),
        ("Windows", Path("tools", "java17")),
        ("Darwin", Path("tools", "java17", "Contents", "Home")),
    ],
)
def test_unparseable_javac_version(mock_tools, host_os, java_home, tmp_path, capsys):
    """If the javac version can't be parsed, the briefcase JDK is used."""
    # Mock os
    mock_tools.host_os = host_os

    # Setup explicit JAVA_HOME
    mock_tools.os.environ = {"JAVA_HOME": "/path/to/nowhere"}

    # Mock return value from javac.
    mock_tools.subprocess.check_output.return_value = "NONSENSE\n"

    # Create a directory to make it look like the Briefcase Java already exists.
    (tmp_path / java_home / "bin").mkdir(parents=True)

    # Create a JDK wrapper by verification
    jdk = JDK.verify(mock_tools)

    # The JDK should have the briefcase JAVA_HOME
    assert jdk.java_home == tmp_path / java_home

    # A single call was made to check javac
    mock_tools.subprocess.check_output.assert_called_once_with(
        [Path("/path/to/nowhere/bin/javac"), "-version"],
    ),

    # No console output (because Briefcase JDK exists)
    output = capsys.readouterr()
    assert (
        "WARNING: Unable to determine the version of Java that is installed"
        in output.out
    )
    assert output.out.endswith("****\n")
    assert output.err == ""


@pytest.mark.parametrize(
    "host_os, host_arch, jdk_url, jhome, is_32bit",
    [
        (
            "Darwin",
            "x86_64",
            "https://github.com/adoptium/temurin17-binaries/releases/download/"
            f"jdk-{JDK_RELEASE}+{JDK_BUILD}/OpenJDK17U-jdk_x64_mac_hotspot_{JDK_RELEASE}_{JDK_BUILD}.tar.gz",
            "java17/Contents/Home",
            False,
        ),
        (
            "Darwin",
            "arm64",
            "https://github.com/adoptium/temurin17-binaries/releases/download/"
            f"jdk-{JDK_RELEASE}+{JDK_BUILD}/OpenJDK17U-jdk_aarch64_mac_hotspot_{JDK_RELEASE}_{JDK_BUILD}.tar.gz",
            "java17/Contents/Home",
            False,
        ),
        (
            "Linux",
            "x86_64",
            "https://github.com/adoptium/temurin17-binaries/releases/download/"
            f"jdk-{JDK_RELEASE}+{JDK_BUILD}/OpenJDK17U-jdk_x64_linux_hotspot_{JDK_RELEASE}_{JDK_BUILD}.tar.gz",
            "java17",
            False,
        ),
        (
            "Linux",
            "aarch64",
            "https://github.com/adoptium/temurin17-binaries/releases/download/"
            f"jdk-{JDK_RELEASE}+{JDK_BUILD}/OpenJDK17U-jdk_aarch64_linux_hotspot_{JDK_RELEASE}_{JDK_BUILD}.tar.gz",
            "java17",
            False,
        ),
        (
            "Linux",
            "aarch64",
            "https://github.com/adoptium/temurin17-binaries/releases/download/"
            f"jdk-{JDK_RELEASE}+{JDK_BUILD}/OpenJDK17U-jdk_arm_linux_hotspot_{JDK_RELEASE}_{JDK_BUILD}.tar.gz",
            "java17",
            True,
        ),
        (
            "Linux",
            "armv7l",
            "https://github.com/adoptium/temurin17-binaries/releases/download/"
            f"jdk-{JDK_RELEASE}+{JDK_BUILD}/OpenJDK17U-jdk_arm_linux_hotspot_{JDK_RELEASE}_{JDK_BUILD}.tar.gz",
            "java17",
            False,
        ),
        (
            "Linux",
            "armv8l",
            "https://github.com/adoptium/temurin17-binaries/releases/download/"
            f"jdk-{JDK_RELEASE}+{JDK_BUILD}/OpenJDK17U-jdk_arm_linux_hotspot_{JDK_RELEASE}_{JDK_BUILD}.tar.gz",
            "java17",
            False,
        ),
        (
            "Windows",
            "AMD64",
            "https://github.com/adoptium/temurin17-binaries/releases/download/"
            f"jdk-{JDK_RELEASE}+{JDK_BUILD}/OpenJDK17U-jdk_x64_windows_hotspot_{JDK_RELEASE}_{JDK_BUILD}.zip",
            "java17",
            False,
        ),
    ],
)
def test_successful_jdk_download(
    mock_tools,
    tmp_path,
    capsys,
    host_os,
    host_arch,
    jdk_url,
    jhome,
    is_32bit,
):
    """If needed, a JDK can be downloaded."""
    # Mock host OS and arch
    mock_tools.host_os = host_os
    mock_tools.host_arch = host_arch
    mock_tools.is_32bit_python = is_32bit

    # Mock a JAVA_HOME that won't exist
    # This is only needed to make macOS *not* run /usr/libexec/java_home
    mock_tools.os.environ = {"JAVA_HOME": "/does/not/exist"}

    # Mock the cached download path
    jdk_zip_path = create_zip_file(tmp_path / "download.zip", content=[("jdk", "jdk")])
    mock_tools.file.download = MagicMock(return_value=jdk_zip_path)

    # Create a directory to make it look like Java was downloaded and unpacked.
    (tmp_path / "tools" / f"jdk-{JDK_RELEASE}+{JDK_BUILD}").mkdir(parents=True)

    # Invoke the verify call
    JDK.verify(mock_tools)

    assert mock_tools.java.java_home == tmp_path / "tools" / jhome

    # Console output contains a warning about the bad JDK location
    output = capsys.readouterr()
    assert output.err == ""
    assert "** WARNING: JAVA_HOME does not point to a Java 17 JDK" in output.out

    # Download was invoked
    mock_tools.file.download.assert_called_with(
        url=jdk_url,
        download_path=tmp_path / "tools",
        role="Java 17 JDK",
    )
    # The archive was unpacked
    mock_tools.shutil.unpack_archive.assert_called_with(
        filename=os.fsdecode(jdk_zip_path), extract_dir=os.fsdecode(tmp_path / "tools")
    )
    # The original archive was deleted
    assert not jdk_zip_path.exists()
    # The download URL for JDK exists
    assert_url_resolvable(mock_tools.java.OpenJDK_download_url)


def test_not_installed(mock_tools, tmp_path):
    """If the JDK isn't installed, and install isn't requested, an error is raised."""
    # Mock host OS
    mock_tools.host_os = "Linux"

    # Invoke the verify call. Install is not requested, so this will fail.
    with pytest.raises(MissingToolError):
        JDK.verify(mock_tools, install=False)

    # Download was not invoked
    assert mock_tools.file.download.call_count == 0


def test_jdk_download_failure(mock_tools, tmp_path):
    """If an error occurs downloading the JDK, an error is raised."""
    # Mock Linux x86_64 as the host
    mock_tools.host_os = "Linux"
    mock_tools.host_arch = "x86_64"

    # Mock a failure on download
    mock_tools.file.download = MagicMock(side_effect=NetworkFailure("mock"))

    # Invoking verify_jdk causes a network failure.
    with pytest.raises(NetworkFailure, match="Unable to mock"):
        JDK.verify(mock_tools)

    # That download was attempted
    mock_tools.file.download.assert_called_with(
        url="https://github.com/adoptium/temurin17-binaries/releases/download/"
        f"jdk-{JDK_RELEASE}+{JDK_BUILD}/OpenJDK17U-jdk_x64_linux_hotspot_{JDK_RELEASE}_{JDK_BUILD}.tar.gz",
        download_path=tmp_path / "tools",
        role="Java 17 JDK",
    )
    # No attempt was made to unpack the archive
    assert mock_tools.shutil.unpack_archive.call_count == 0


def test_invalid_jdk_archive(mock_tools, tmp_path):
    """If the JDK download isn't a valid archive, raise an error."""
    # Mock Linux x86_64 as the host
    mock_tools.host_os = "Linux"
    mock_tools.host_arch = "x86_64"

    # Mock the cached download path
    jdk_zip_path = create_zip_file(tmp_path / "download.zip", content=[("jdk", "jdk")])
    mock_tools.file.download = MagicMock(return_value=jdk_zip_path)

    # Mock an unpack failure due to an invalid archive
    mock_tools.shutil.unpack_archive.side_effect = shutil.ReadError

    with pytest.raises(BriefcaseCommandError):
        JDK.verify(mock_tools)

    # The download occurred
    mock_tools.file.download.assert_called_with(
        url="https://github.com/adoptium/temurin17-binaries/releases/download/"
        f"jdk-{JDK_RELEASE}+{JDK_BUILD}/OpenJDK17U-jdk_x64_linux_hotspot_{JDK_RELEASE}_{JDK_BUILD}.tar.gz",
        download_path=tmp_path / "tools",
        role="Java 17 JDK",
    )
    # An attempt was made to unpack the archive.
    mock_tools.shutil.unpack_archive.assert_called_with(
        filename=os.fsdecode(jdk_zip_path),
        extract_dir=os.fsdecode(tmp_path / "tools"),
    )
    # The original archive was not deleted
    assert jdk_zip_path.exists()



import gzip
import os
import subprocess
import sys
from unittest import mock

import pytest

from briefcase.console import Console, Log
from briefcase.exceptions import BriefcaseCommandError
from briefcase.platforms.linux.system import LinuxSystemBuildCommand

from ....utils import create_file


@pytest.fixture
def build_command(tmp_path, first_app):
    command = LinuxSystemBuildCommand(
        logger=Log(),
        console=Console(),
        base_path=tmp_path / "base_path",
        data_path=tmp_path / "briefcase",
        apps={"first": first_app},
    )
    command.tools.host_os = "Linux"
    command.tools.host_arch = "wonky"

    # Mock subprocess
    command.tools.subprocess = mock.MagicMock()

    # Mock the app context
    command.tools.app_tools[first_app].app_context = mock.MagicMock()

    return command


@pytest.mark.skipif(sys.platform == "win32", reason="Can't build Linux apps on Windows")
def test_build_app(build_command, first_app, tmp_path):
    """An app can be built as a deb."""
    # Build the app
    build_command.build_app(first_app)

    # The bootstrap binary was compiled
    bundle_path = tmp_path / "base_path/build/first-app/somevendor/surprising"
    build_command.tools[first_app].app_context.run.assert_called_with(
        ["make", "-C", "bootstrap", "install"],
        check=True,
        cwd=bundle_path,
    )

    # The license file has been installed
    doc_path = bundle_path / "first-app-0.0.1/usr/share/doc/first-app"
    assert (doc_path / "copyright").exists()
    with (doc_path / "copyright").open(encoding="utf-8") as f:
        assert f.read() == "First App License"

    # The Changelog has been compressed and installed
    assert (doc_path / "changelog.gz").exists()
    with gzip.open(doc_path / "changelog.gz") as f:
        assert f.read().decode() == "First App Changelog"

    # The manpage has been installed
    man_path = bundle_path / "first-app-0.0.1/usr/share/man/man1"
    assert (man_path / "first-app.1.gz").exists()
    with gzip.open(man_path / "first-app.1.gz") as f:
        assert f.read().decode() == "First App manpage"

    # Problematic permissions have been updated
    lib_dir = bundle_path / "first-app-0.0.1/usr/lib/first-app"
    # 775 -> 775
    assert os.stat(lib_dir / "app/support.so").st_mode & 0o777 == 0o755
    # 664 -> 644
    assert (
        os.stat(lib_dir / "app_packages/secondlib/second_a.so").st_mode & 0o777 == 0o644
    )
    # no perms change
    assert os.stat(lib_dir / "app/support_same_perms.so").st_mode & 0o777 == 0o744

    # Strip has been invoked on the binary
    build_command.tools.subprocess.check_output.assert_called_once_with(
        [
            "strip",
            bundle_path / "first-app-0.0.1/usr/bin/first-app",
        ]
    )


def test_build_bootstrap_failed(build_command, first_app, tmp_path):
    """If the bootstrap binary can't be compiled, an error is raised."""
    # Mock a build failure
    build_command.tools[first_app].app_context.run.side_effect = (
        subprocess.CalledProcessError(cmd=["make ..."], returncode=-1)
    )

    # Build the app; it will fail
    with pytest.raises(
        BriefcaseCommandError,
        match=r"Error building bootstrap binary for first-app.",
    ):
        build_command.build_app(first_app)

    # An attempt to do the compile occurred.
    bundle_path = tmp_path / "base_path/build/first-app/somevendor/surprising"
    build_command.tools[first_app].app_context.run.assert_called_with(
        ["make", "-C", "bootstrap", "install"],
        check=True,
        cwd=bundle_path,
    )


def test_missing_license(build_command, first_app, tmp_path):
    """If the license source file is missing, an error is raised."""
    bundle_path = tmp_path / "base_path/build/first-app/somevendor/surprising"

    # Delete the license source
    (tmp_path / "base_path/LICENSE").unlink()

    # Build the app; it will fail
    with pytest.raises(
        BriefcaseCommandError,
        match=r"Your `pyproject.toml` specifies a license file of 'LICENSE'",
    ):
        build_command.build_app(first_app)

    # The bootstrap binary was compiled
    build_command.tools[first_app].app_context.run.assert_called_with(
        ["make", "-C", "bootstrap", "install"],
        check=True,
        cwd=bundle_path,
    )


def test_specified_license_file_is_copied(build_command, first_app, tmp_path):
    """The specified license file is copied if a license file is specified."""
    create_file(tmp_path / "base_path/LICENSE.txt", "The Actual First App License")
    first_app.license["file"] = "LICENSE.txt"

    # Build the app
    build_command.build_app(first_app)

    # The correct license file has been copied
    doc_folder = (
        build_command.bundle_path(first_app)
        / f"{first_app.app_name}-{first_app.version}"
        / "usr"
        / "share"
        / "doc"
        / first_app.app_name
    )
    assert (doc_folder / "copyright").read_text(
        encoding="utf-8"
    ) == "The Actual First App License"


def test_license_text_is_saved(build_command, first_app):
    """The license text is saved as a file in the bundle."""
    first_app.license = {"text": "Some license text"}

    # Build the app
    build_command.build_app(first_app)

    # The license text has been saved
    doc_folder = (
        build_command.bundle_path(first_app)
        / f"{first_app.app_name}-{first_app.version}"
        / "usr"
        / "share"
        / "doc"
        / first_app.app_name
    )

    assert (doc_folder / "copyright").read_text(encoding="utf-8") == "Some license text"


def test_license_text_warns_with_single_line_license(build_command, first_app):
    """A warning is logged if the license text is a single line."""

    first_app.license = {"text": "Some license text"}
    build_command.logger.warning = mock.MagicMock()

    # Build the app
    build_command.build_app(first_app)

    build_command.logger.warning.assert_called_once_with(
        """
Your app specifies a license using `license.text`, but the value doesn't appear to be a
full license. Briefcase will generate a `copyright` file for your project; you should
ensure that the contents of this file is adequate.
"""
    )


def test_exception_with_no_license(build_command, first_app):
    """An exception is raised if there is no license defined."""

    first_app.license = {}
    build_command.logger.warning = mock.MagicMock()

    # Build the app
    with pytest.raises(
        BriefcaseCommandError,
        match="Your project does not contain a LICENSE definition.",
    ):
        build_command.build_app(first_app)


def test_license_text_doesnt_warn_with_multi_line_license(
    build_command, first_app, tmp_path
):
    """No warning is logged if the license text is multi-line."""
    first_app.license = {"text": "Some license text\nsome more text"}
    build_command.logger.warning = mock.MagicMock()

    # Build the app
    build_command.build_app(first_app)
    build_command.logger.warning.assert_not_called()


def test_missing_changelog(build_command, first_app, tmp_path):
    """If the changelog source file is missing, an error is raised."""
    bundle_path = tmp_path / "base_path/build/first-app/somevendor/surprising"

    # Delete the changelog source
    (tmp_path / "base_path/CHANGELOG").unlink()

    # Build the app; it will fail
    with pytest.raises(
        BriefcaseCommandError,
        match=r"Your project does not contain a CHANGELOG file.",
    ):
        build_command.build_app(first_app)

    # The bootstrap binary was compiled
    build_command.tools[first_app].app_context.run.assert_called_with(
        ["make", "-C", "bootstrap", "install"],
        check=True,
        cwd=bundle_path,
    )

    # The license file has been installed
    doc_path = bundle_path / "first-app-0.0.1/usr/share/doc/first-app"
    assert (doc_path / "copyright").exists()
    with (doc_path / "copyright").open(encoding="utf-8") as f:
        assert f.read() == "First App License"


def test_missing_manpage(build_command, first_app, tmp_path):
    """If the manpage source file is missing, an error is raised."""
    bundle_path = tmp_path / "base_path/build/first-app/somevendor/surprising"

    # Delete the manpage source
    (bundle_path / "first-app.1").unlink()

    # Build the app; it will fail
    with pytest.raises(
        BriefcaseCommandError,
        match=r"Template does not provide a manpage source file `first-app\.1`",
    ):
        build_command.build_app(first_app)

    # The bootstrap binary was compiled
    build_command.tools[first_app].app_context.run.assert_called_with(
        ["make", "-C", "bootstrap", "install"],
        check=True,
        cwd=bundle_path,
    )

    # The license file has been installed
    doc_path = bundle_path / "first-app-0.0.1/usr/share/doc/first-app"
    assert (doc_path / "copyright").exists()
    with (doc_path / "copyright").open(encoding="utf-8") as f:
        assert f.read() == "First App License"

    # The Changelog has been compressed and installed
    assert (doc_path / "changelog.gz").exists()
    with gzip.open(doc_path / "changelog.gz") as f:
        assert f.read().decode() == "First App Changelog"



import pytest

from ...utils import create_file


def test_valid_test_source_dir(convert_command):
    """Valid test_source_dir raises no errors and returns True."""
    app_name = "newtarget"
    test_source_dir = "tests"

    full_test_source_dir = convert_command.base_path / test_source_dir
    full_test_source_dir.mkdir(parents=True)

    assert convert_command.validate_test_source_dir(app_name, test_source_dir)


def test_test_source_dir_shouldnt_contain_test_entry_script(convert_command):
    """If `convert_command.base_path/test_source_dir` already contains a test entry
    script, a ValueError is raised."""
    app_name = "newtarget"
    test_source_dir = "tests"

    test_entry_path = convert_command.base_path / test_source_dir / f"{app_name}.py"
    create_file(test_entry_path, "TEST ENTRY SCRIPT")

    with pytest.raises(ValueError):
        convert_command.validate_test_source_dir(app_name, test_source_dir)



def test_packaging_formats(package_command):
    assert package_command.packaging_formats == ["aab", "apk", "debug-apk"]


def test_default_packaging_format(package_command):
    assert package_command.default_packaging_format == "aab"



import pytest


@pytest.mark.parametrize(
    "name",
    [
        "helloworld",
        "helloWorld",
        "hello42world",
        "42helloworld",  # ?? Are we sure this is correct?
        "hello_world",
        "hello-world",
    ],
)
def test_valid_app_name(new_command, name):
    """Test that valid app names are accepted."""
    assert new_command.validate_app_name(name)


@pytest.mark.parametrize(
    "name",
    [
        "hello world",  # contains a space
        "helloworld!",  # contains punctuation
        "_helloworld",  # leading underscore
        "-helloworld",  # leading hyphen
        "pass",  # python keyword
        "Pass",  # python keyword, but different case usage
        "PASS",  # python keyword, but all upper case
        "in",  # javascript keyword
        "In",  # javascript keyword, but different case usage
        "IN",  # javascript keyword, but all upper case
        "synchronized",  # Java keyword
        "Synchronized",  # Java keyword, but different case usage
        "SYNCHRONIZED",  # Java keyword, but all upper case
        "false",  # Python, Java and Javascript keyword (in different cases)
        "False",  # Python, Java and Javascript keyword (in different cases)
        "FALSE",  # Python, Java and Javascript keyword (in different cases)
        "existing",  # pre-existing directory name
    ],
)
def test_invalid_app_name(new_command, name, tmp_path):
    """Test that invalid app names are rejected."""
    (tmp_path / "existing").mkdir()

    with pytest.raises(ValueError):
        new_command.validate_app_name(name)



import pytest

from briefcase.exceptions import BriefcaseCommandError

from ...utils import create_file


def test_no_target_versions(base_command, my_app):
    """Verification succeeds if no target versions are declared."""
    base_command.platform_target_version = None

    create_file(base_command.bundle_path(my_app) / "briefcase.toml", content="")

    base_command.verify_app_template(my_app)


def test_platform_version_compatible(base_command, my_app):
    """Verification succeeds when template version matches platform version."""
    base_command.platform_target_version = "42.42"

    create_file(
        base_command.bundle_path(my_app) / "briefcase.toml",
        content="[briefcase]\ntarget_version = '42.42'",
    )

    base_command.verify_app_template(my_app)


@pytest.mark.parametrize("template_version", ["", "32.32", "52.52"])
def test_platform_version_incompatible(base_command, my_app, template_version):
    """Verification fails when template version doesn't match platform version."""
    base_command.platform_target_version = "42.42"

    create_file(
        base_command.bundle_path(my_app) / "briefcase.toml",
        content=f"[briefcase]\ntarget_version = '{template_version}'",
    )

    with pytest.raises(
        BriefcaseCommandError,
        match="The app template used to generate this app is not compatible with this version\nof Briefcase.",
    ):
        base_command.verify_app_template(my_app)



import pytest

from briefcase.commands.base import parse_config_overrides
from briefcase.exceptions import BriefcaseConfigError


@pytest.mark.parametrize(
    "overrides, values",
    [
        # No content
        (None, {}),
        ([], {}),
        # Boolean
        (["key=true"], {"key": True}),
        # Integers
        (["key=42"], {"key": 42}),
        (["key=-42"], {"key": -42}),
        # Integers
        (["key=42.37"], {"key": 42.37}),
        (["key=-42.37"], {"key": -42.37}),
        # Strings
        (["key='hello'"], {"key": "hello"}),
        (["key=''"], {"key": ""}),
        (["key='42'"], {"key": "42"}),
        (['key="hello"'], {"key": "hello"}),
        (['key=""'], {"key": ""}),
        (['key="42"'], {"key": "42"}),
        # List
        (['key=[1, "two", true]'], {"key": [1, "two", True]}),
        # Dictionary
        (['key={a=1, b="two", c=true}'], {"key": {"a": 1, "b": "two", "c": True}}),
        # Multiple values
        (
            [
                "key1=42",
                'key2="hello"',
                'key3=[1, "two", true]',
                'key4={a=1, b="two", c=true}',
            ],
            {
                "key1": 42,
                "key2": "hello",
                "key3": [1, "two", True],
                "key4": {"a": 1, "b": "two", "c": True},
            },
        ),
    ],
)
def test_valid_overrides(overrides, values):
    """Valid values can be parsed as config overrides."""
    assert parse_config_overrides(overrides) == values


@pytest.mark.parametrize(
    "overrides, message",
    [
        # Bare string
        (["foobar"], r"Unable to parse configuration override "),
        # Unquoted string
        (["key=foobar"], r"Unable to parse configuration override "),
        # Unbalanced quote
        (["key='foobar"], r"Unable to parse configuration override "),
        # Unmatched brackets
        (['key=[1, "two",'], r"Unable to parse configuration override "),
        # Unmatches parentheses
        (['key={a=1, b="two"'], r"Unable to parse configuration override "),
        # Valid value followed by invalid
        (["good=42", "key=foobar"], r"Unable to parse configuration override "),
        # Space in the key.
        (["spacy key=42"], r"Unable to parse configuration override "),
        # Multi-level key. This is legal TOML, but difficult to merge.
        (["multi.level.key=42"], r"Can't override multi-level configuration keys\."),
        # Key that can't be overridden
        (["app_name='foobar'"], r"The app name cannot be overridden\."),
    ],
)
def test_invalid_overrides(overrides, message):
    with pytest.raises(BriefcaseConfigError, match=message):
        parse_config_overrides(overrides)



import pytest

from briefcase.exceptions import BriefcaseCommandError


def test_unvalidated_input(new_command):
    """If the user enters text and there's no validation, the text is returned."""
    new_command.input.values = ["hello"]

    value = new_command.input_text(
        intro="Some introduction", variable="my variable", default="goodbye"
    )

    assert new_command.input.prompts == ["My Variable [goodbye]: "]
    assert value == "hello"


def test_unvalidated_input_with_override(new_command):
    """If an override is provided and there's no validation, the override is
    returned."""
    new_command.input.values = ["hello"]

    value = new_command.input_text(
        intro="Some introduction",
        variable="my variable",
        default="goodbye",
        override_value="override",
    )

    assert new_command.input.prompts == []
    assert value == "override"


def test_validated_input(new_command):
    """If the user enters text and there's validation, the user is prompted until valid
    text is entered."""
    new_command.input.values = ["bad", "hello"]

    def validator(text):
        if text == "bad":
            raise ValueError("That's bad...")
        return True

    value = new_command.input_text(
        intro="Some introduction",
        variable="my variable",
        default="goodbye",
        validator=validator,
    )

    assert new_command.input.prompts == [
        "My Variable [goodbye]: ",
        "My Variable [goodbye]: ",
    ]
    assert value == "hello"


def test_validated_input_with_override(new_command):
    """If an override is provided and there's validation, the override is validated and
    returned."""

    def validator(text):
        if text == "bad":
            raise ValueError("That's bad...")
        return True

    value = new_command.input_text(
        intro="Some introduction",
        variable="my variable",
        default="goodbye",
        validator=validator,
        override_value="override",
    )

    assert new_command.input.prompts == []
    assert value == "override"


def test_validated_input_with_invalid_override(new_command):
    """If an invalid override is provided and there's validation, the override is
    rejected and the user is prompted until valid text is entered."""
    new_command.input.values = ["bad", "hello"]

    def validator(text):
        if text in {"bad", "bad-override"}:
            return False
        return True

    value = new_command.input_text(
        intro="Some introduction",
        variable="my variable",
        default="goodbye",
        validator=validator,
        override_value="bad-override",
    )

    assert new_command.input.prompts == [
        "My Variable [goodbye]: ",
        "My Variable [goodbye]: ",
    ]
    assert value == "hello"


def test_input_with_default(new_command):
    """If the user enters text and there's no validation, the text is returned."""
    new_command.input.values = [""]

    value = new_command.input_text(
        intro="Some introduction", variable="my variable", default="goodbye"
    )

    assert new_command.input.prompts == ["My Variable [goodbye]: "]
    assert value == "goodbye"


def test_input_disabled(new_command):
    """If input is disabled, the default is returned."""
    new_command.input.enabled = False

    value = new_command.input_text(
        intro="Some introduction",
        variable="my variable",
        default="goodbye",
    )

    assert new_command.input.prompts == []
    assert value == "goodbye"


def test_input_disabled_with_override(new_command):
    """If input is disabled, the override is returned."""
    new_command.input.enabled = False

    value = new_command.input_text(
        intro="Some introduction",
        variable="my variable",
        default="goodbye",
        override_value="override",
    )

    assert new_command.input.prompts == []
    assert value == "override"


def test_input_disabled_validation_failure(new_command):
    """If input is disabled, and validation fails, an error is raised."""
    new_command.input.enabled = False

    with pytest.raises(BriefcaseCommandError, match="Well that won't work..."):

        def not_valid(text):
            raise ValueError("Well that won't work...")

        new_command.input_text(
            intro="Some introduction",
            variable="my variable",
            default="goodbye",
            validator=not_valid,
        )

    assert new_command.input.prompts == []


def test_input_disabled_validation_failure_with_override(new_command):
    """If input is disabled, and validation fails for override, an error is raised."""
    new_command.input.enabled = False

    with pytest.raises(BriefcaseCommandError, match="Well that won't work..."):

        def not_valid(text):
            raise ValueError("Well that won't work...")

        new_command.input_text(
            intro="Some introduction",
            variable="my variable",
            default="goodbye",
            validator=not_valid,
            override_value="override",
        )

    assert new_command.input.prompts == []


def test_prompt_capitalization(new_command):
    """The prompt is capitalized appropriately."""
    new_command.input.values = ["hello"]

    new_command.input_text(
        intro="Some introduction", variable="user's URL", default="goodbye"
    )

    assert new_command.input.prompts == ["User's URL [goodbye]: "]



def test_avd_config(mock_tools, android_sdk, tmp_path):
    """An AVD configuration can be read."""
    config_file = tmp_path / "home/.android/avd/testDevice.avd/config.ini"
    config_file.parent.mkdir(parents=True)

    # Write a default config. It contains:
    # * blank lines
    # * a key whose value explicitly contains an equals sign.
    with config_file.open("w", encoding="utf-8") as f:
        f.write(
            """
avd.ini.encoding=UTF-8
hw.device.manufacturer=Google
hw.device.name=pixel
weird.key=good=bad

PlayStore.enabled=no
avd.name=testDevice
disk.cachePartition=yes
disk.cachePartition.size=37MB
"""
        )

    assert android_sdk.avd_config("testDevice") == {
        "avd.ini.encoding": "UTF-8",
        "hw.device.manufacturer": "Google",
        "hw.device.name": "pixel",
        "weird.key": "good=bad",
        "PlayStore.enabled": "no",
        "avd.name": "testDevice",
        "disk.cachePartition": "yes",
        "disk.cachePartition.size": "37MB",
    }


def test_avd_config_with_space(mock_tools, android_sdk, tmp_path):
    """An AVD configuration that contains spaces can be read."""
    config_file = tmp_path / "home/.android/avd/testDevice.avd/config.ini"
    config_file.parent.mkdir(parents=True)

    # Write a default config. It contains:
    # * blank lines
    # * a key whose value explicitly contains an equals sign.
    # * spaces either side of the key/value separator
    with config_file.open("w", encoding="utf-8") as f:
        f.write(
            """
avd.ini.encoding = UTF-8
hw.device.manufacturer = Google
hw.device.name = pixel
weird.key = good=bad

PlayStore.enabled = no
avd.name = testDevice
disk.cachePartition = yes
disk.cachePartition.size = 37MB
"""
        )

    assert android_sdk.avd_config("testDevice") == {
        "avd.ini.encoding": "UTF-8",
        "hw.device.manufacturer": "Google",
        "hw.device.name": "pixel",
        "weird.key": "good=bad",
        "PlayStore.enabled": "no",
        "avd.name": "testDevice",
        "disk.cachePartition": "yes",
        "disk.cachePartition.size": "37MB",
    }



import pytest

from briefcase.console import InputDisabled
from tests.utils import default_rich_prompt


@pytest.mark.parametrize(
    "user_input, expected",
    [
        ("y", True),
        ("Y", True),
        ("yes", True),
        ("YES", True),
        ("Yes", True),
        ("YeS", True),
        ("Yay", True),
        ("Yaaaas!", True),
        ("n", False),
        ("N", False),
        ("No", False),
        ("NO", False),
        ("Nay", False),
        ("never!", False),
    ],
)
def test_boolean_input(console, user_input, expected):
    question = "Are you handsome"
    prompt = "Are you handsome [y/N]? "
    console.input.side_effect = [user_input]

    result = console.boolean_input(question=question)

    assert result == expected
    console.input.assert_called_once_with(default_rich_prompt(prompt), markup=True)


def test_boolean_default_true(console):
    """If True is the default, it is returned."""
    question = "Are you handsome"
    prompt = "Are you handsome [Y/n]? "
    console.input.side_effect = [""]

    result = console.boolean_input(question=question, default=True)

    assert result
    console.input.assert_called_once_with(default_rich_prompt(prompt), markup=True)


def test_boolean_default_false(console):
    """If False is the default, it is returned."""
    question = "Are you handsome"
    prompt = "Are you handsome [y/N]? "
    console.input.side_effect = [""]

    result = console.boolean_input(question=question, default=False)

    assert not result
    console.input.assert_called_once_with(default_rich_prompt(prompt), markup=True)


def test_boolean_default_None(console):
    """If no default is specified, no response is not accepted."""
    question = "Are you handsome"
    console.input.side_effect = ["", "y"]

    result = console.boolean_input(question=question, default=None)

    assert result
    assert console.input.call_count == 2


def test_bad_input(console):
    question = "Are you handsome"
    console.input.side_effect = ["pork", "ham", "spam", "Yam"]

    result = console.boolean_input(question=question)

    assert result
    assert console.input.call_count == 4


def test_disabled(disabled_console):
    """If input is disabled, the default is returned."""
    question = "Are you handsome "

    result = disabled_console.boolean_input(question=question)

    assert not result
    disabled_console.input.assert_not_called()


def test_disabled_no_default(disabled_console):
    """If input is disabled and there is no default, an error is raised."""
    question = "Are you handsome "

    with pytest.raises(InputDisabled):
        disabled_console.boolean_input(question=question, default=None)

    disabled_console.input.assert_not_called()



import threading
from io import StringIO

import pytest

from briefcase.console import Log
from briefcase.integrations import subprocess
from briefcase.integrations.subprocess import PopenOutputStreamer


@pytest.fixture
def streamer(streaming_process):
    return PopenOutputStreamer(
        label="test",
        popen_process=streaming_process,
        logger=Log(),
    )


def test_output(streamer, capsys):
    """Process output is printed."""
    streamer.start()
    streamer.join(timeout=5)

    # fmt: off
    assert capsys.readouterr().out == (
        "output line 1\n"
        "\n"
        "output line 3\n"
    )
    # fmt: on


def test_stdout_closes_unexpectedly(streamer, monkeypatch, capsys):
    """Streamer exits from ValueError because stdout was closed."""

    def monkeypatch_ensure_str(value):
        """Close stdout when ensure_str() runs on output from readline()."""
        streamer.popen_process.stdout.close()
        return value

    monkeypatch.setattr(subprocess, "ensure_str", monkeypatch_ensure_str)
    streamer.popen_process.stdout = StringIO("output line 1\noutput line 2")

    streamer.start()
    streamer.join(timeout=5)

    assert capsys.readouterr().out == (
        "output line 1\n"
        "WARNING: stdout was unexpectedly closed while streaming output\n"
    )


def test_readline_raises_exception(streamer, monkeypatch, capsys):
    """Streamer aborts if readline() raises ValueError for reasons other than stdout
    closing."""

    def monkeypatch_ensure_str(value):
        """Simulate readline() raising an ValueError-derived exception."""
        raise UnicodeError("readline() exception")

    monkeypatch.setattr(subprocess, "ensure_str", monkeypatch_ensure_str)

    streamer.start()
    streamer.join(timeout=5)

    assert capsys.readouterr().out == (
        "Error while streaming output: UnicodeError: readline() exception\n"
    )


def test_request_stop(streamer, capsys):
    """Requesting the streamer to stop sets the stop flag."""
    streamer.start()
    streamer.join(timeout=5)

    # fmt: off
    assert capsys.readouterr().out == (
        "output line 1\n"
        "\n"
        "output line 3\n"
    )
    # fmt: on

    assert not streamer.stop_flag.is_set()

    streamer.request_stop()

    assert streamer.stop_flag.is_set()


def test_request_stop_set_immediately(streamer, capsys):
    """Nothing is printed if stop flag is immediately set."""
    streamer.request_stop()

    streamer.start()
    streamer.join(timeout=5)

    assert capsys.readouterr().out == ""


def test_request_stop_set_during_output(streamer, monkeypatch, capsys):
    """Streamer prints nothing more after stop flag is set."""

    def filter_func(value):
        """Simulate stop flag set while output is being read."""
        streamer.request_stop()
        yield value

    streamer.filter_func = filter_func

    streamer.start()
    streamer.join(timeout=5)

    assert capsys.readouterr().out == "output line 1\n"


def test_captured_output(streamer, capsys):
    """Captured output is available after streaming."""
    streamer.capture_output = True

    streamer.start()
    streamer.join(timeout=5)

    assert capsys.readouterr().out == ""

    # fmt: off
    assert streamer.captured_output == (
        "output line 1\n"
        "\n"
        "output line 3\n"
    )
    # fmt: on


def test_captured_output_interim(streamer, monkeypatch, capsys):
    """Captured output is available during streaming."""
    streamer.capture_output = True

    streamer_is_waiting = threading.Event()
    continue_streaming = threading.Event()

    def mock_readline():
        """Wait on asserts while returning output lines."""
        yield "output line 1\n"
        yield "output line 2\n"
        streamer_is_waiting.set()
        continue_streaming.wait(timeout=5)
        yield "output line 3\n"
        yield "output line 4\n"
        yield ""

    streamer.popen_process.stdout.readline.side_effect = mock_readline()

    streamer.start()
    streamer_is_waiting.wait(timeout=5)

    assert streamer.captured_output == "output line 1\noutput line 2\n"

    continue_streaming.set()
    streamer.join(timeout=5)

    assert streamer.captured_output == "output line 3\noutput line 4\n"


def test_filter_func(streamer, capsys):
    """A filter can be added to modify an output stream."""

    # Define a filter function that converts "output" into "filtered"
    def filter_func(line):
        yield line.replace("output", "filtered")

    streamer.filter_func = filter_func

    streamer.start()
    streamer.join(timeout=5)

    # fmt: off
    # Output has been transformed
    assert capsys.readouterr().out == (
        "filtered line 1\n"
        "\n"
        "filtered line 3\n"
    )
    # fmt: on


def test_filter_func_reject(streamer, capsys):
    """A filter that rejects lines can be added to modify an output stream."""

    # Define a filter function that ignores blank lines
    def filter_func(line):
        if len(line) == 0:
            return
        yield line

    streamer.filter_func = filter_func

    streamer.start()
    streamer.join(timeout=5)

    # fmt: off
    # Output has been transformed
    assert capsys.readouterr().out == (
        "output line 1\n"
        "output line 3\n"
    )
    # fmt: on


def test_filter_func_line_ends(streamer, sleep_zero, capsys):
    """Filter functions are not provided the newline."""

    # Define a filter function that redacts lines that end with 1
    # The newline is *not* included.
    def filter_func(line):
        if line.endswith("line 1"):
            yield line.replace("line 1", "**REDACTED**")
        else:
            yield line

    streamer.filter_func = filter_func

    streamer.start()
    streamer.join(timeout=5)

    # fmt: off
    # Output has been transformed; newline exists in output
    assert capsys.readouterr().out == (
        "output **REDACTED**\n"
        "\n"
        "output line 3\n"
    )
    # fmt: on


def test_filter_func_line_multiple_output(streamer, capsys):
    """Filter functions can generate multiple lines from a single input."""

    # Define a filter function that adds an extra line of content when the
    # lines that end with 1
    def filter_func(line):
        yield line
        if line.endswith("line 1"):
            yield "Extra content!"

    streamer.filter_func = filter_func

    streamer.start()
    streamer.join(timeout=5)

    # fmt: off
    # Output has been transformed; newline exists in output
    assert capsys.readouterr().out == (
        "output line 1\n"
        "Extra content!\n"
        "\n"
        "output line 3\n"
    )
    # fmt: on


def test_filter_func_stop_iteration(streamer, capsys):
    """A filter can indicate that logging should stop."""

    # Define a filter function that converts "output" into "filtered",
    # and terminates streaming when a blank line is seen.
    def filter_func(line):
        if line == "":
            raise subprocess.StopStreaming()
        yield line.replace("output", "filtered")

    streamer.filter_func = filter_func

    streamer.start()
    streamer.join(timeout=5)

    # fmt: off
    # Output has been transformed, but is truncated when the empty line was received.
    assert capsys.readouterr().out == (
        "filtered line 1\n"
    )
    # fmt: on


def test_filter_func_output_and_stop_iteration(streamer, capsys):
    """A filter can indicate that logging should stop, and also output content."""

    # Define a filter function that converts "output" into "filtered",
    # and terminates streaming when a blank line is seen; but outputs
    # one more line before terminating.
    def filter_func(line):
        if line == "":
            yield "This should be the last line"
            raise subprocess.StopStreaming()
        yield line.replace("output", "filtered")

    streamer.filter_func = filter_func

    streamer.start()
    streamer.join(timeout=5)

    # fmt: off
    # Output has been transformed, but is truncated when the empty line was received.
    assert capsys.readouterr().out == (
        "filtered line 1\n"
        "This should be the last line\n"
    )
    # fmt: on


def test_filter_func_line_unexpected_error(streamer, capsys):
    """If a filter function fails, the error is caught and logged."""

    # Define a filter function that raises a RunTimeError
    # The newline is *not* included.
    def filter_func(line):
        if not line:
            raise RuntimeError("Like something totally went wrong")
        yield line

    streamer.filter_func = filter_func

    streamer.start()
    streamer.join(timeout=5)

    # fmt: off
    # Exception
    assert capsys.readouterr().out == (
        "output line 1\n"
        "Error while streaming output: RuntimeError: Like something totally went wrong\n"
    )
    # fmt: on



from .build import BuildCommand  # noqa: F401
from .convert import ConvertCommand  # noqa: F401
from .create import CreateCommand  # noqa: F401
from .dev import DevCommand  # noqa: F401
from .new import NewCommand  # noqa: F401
from .open import OpenCommand  # noqa: F401
from .package import PackageCommand  # noqa: F401
from .publish import PublishCommand  # noqa: F401
from .run import RunCommand  # noqa: F401
from .update import UpdateCommand  # noqa: F401
from .upgrade import UpgradeCommand  # noqa: F401



import os
import sys
from unittest.mock import MagicMock, PropertyMock

import pytest
import requests

from briefcase.console import Console, Log
from briefcase.integrations.android_sdk import AndroidSDK
from briefcase.integrations.subprocess import Subprocess
from briefcase.platforms.android.gradle import GradlePackageCommand

from ....utils import create_file


@pytest.fixture
def package_command(tmp_path, first_app_config, monkeypatch):
    command = GradlePackageCommand(
        logger=Log(),
        console=Console(),
        base_path=tmp_path / "base_path",
        data_path=tmp_path / "briefcase",
    )
    command.tools.android_sdk = MagicMock(spec_set=AndroidSDK)
    command.tools.os = MagicMock(spec_set=os)
    command.tools.os.environ = {}
    command.tools.sys = MagicMock(spec_set=sys)
    command.tools.requests = MagicMock(spec_set=requests)
    command.tools.subprocess = MagicMock(spec_set=Subprocess)
    monkeypatch.setattr(
        type(command.tools), "system_encoding", PropertyMock(return_value="ISO-42")
    )

    # Make sure the dist folder exists
    (tmp_path / "base_path/dist").mkdir(parents=True)
    return command


@pytest.fixture
def first_app_generated(first_app_config, tmp_path):
    # Create the briefcase.toml file
    create_file(
        tmp_path
        / "base_path"
        / "build"
        / "first-app"
        / "android"
        / "gradle"
        / "briefcase.toml",
        """
[paths]
app_packages_path="app_packages"
support_path="support"
metadata_resource_path="res/briefcase.xml"
""",
    )

    create_file(
        tmp_path
        / "base_path"
        / "build"
        / "first-app"
        / "android"
        / "gradle"
        / "res"
        / "briefcase.xml",
        """<resources></resources>""",
    )
    return first_app_config



import pytest

from briefcase.platforms.macOS.filters import macOS_log_clean_filter


@pytest.mark.parametrize(
    "original, filtered",
    [
        # macOS Logging preamble
        (
            "Filtering the log data using "
            '"senderImagePath == "/path/to/My App.app/Contents/MacOS/My App" '
            'OR (processImagePath == "/path/to/My App.app/Contents/MacOS/My App" '
            'AND senderImagePath == "/usr/lib/libffi.dylib")"',
            None,
        ),
        (
            "Timestamp               Ty Process[PID:TID]",
            None,
        ),
        # iOS Logging preamble
        (
            "Error from getpwuid_r: 0 (Undefined error: 0)",
            None,
        ),
        (
            'Filtering the log data using "senderImagePath ENDSWITH "/Toga Test!" '
            'OR (processImagePath ENDSWITH "/Toga Test!" '
            'AND (senderImagePath ENDSWITH "-iphonesimulator.so" '
            'OR senderImagePath ENDSWITH "-iphonesimulator.dylib"))"',
            None,
        ),
        # Startup log
        (
            "2022-11-14 13:21:14.972 Df My App[59972:780a15] Configure argc/argv...",
            ("Configure argc/argv...", False),
        ),
        # Empty startup log
        (
            "2022-11-14 13:21:14.972 Df My App[59972:780a15] ",
            ("", False),
        ),
        # macOS App log
        (
            "2022-11-14 13:21:15.341 Df My App[59972:780a15] (libffi.dylib) Hello World!",
            ("Hello World!", True),
        ),
        # Empty macOS App log
        (
            "2022-11-14 13:21:15.341 Df My App[59972:780a15] (libffi.dylib) ",
            ("", True),
        ),
        # iOS App log (old style .so libraries)
        (
            "2022-11-14 13:21:15.341 Df My App[59972:780a15] (_ctypes.cpython-312-iphonesimulator.so) Hello World!",
            ("Hello World!", True),
        ),
        (
            "2022-11-14 13:21:15.341 Df My App[59972:780a15] (_ctypes.cpython-38-iphonesimulator.so) Hello World!",
            ("Hello World!", True),
        ),
        # iOS App log
        (
            "2022-11-14 13:21:15.341 Df My App[59972:780a15] (_ctypes.cpython-312-iphonesimulator.dylib) Hello World!",
            ("Hello World!", True),
        ),
        (
            "2022-11-14 13:21:15.341 Df My App[59972:780a15] (_ctypes.cpython-38-iphonesimulator.dylib) Hello World!",
            ("Hello World!", True),
        ),
        # Empty iOS App log (old style .so binaries)
        (
            "2022-11-14 13:21:15.341 Df My App[59972:780a15] (_ctypes.cpython-312-iphonesimulator.so) ",
            ("", True),
        ),
        (
            "2022-11-14 13:21:15.341 Df My App[59972:780a15] (_ctypes.cpython-38-iphonesimulator.so) ",
            ("", True),
        ),
        # Empty iOS App log
        (
            "2022-11-14 13:21:15.341 Df My App[59972:780a15] (_ctypes.cpython-312-iphonesimulator.dylib) ",
            ("", True),
        ),
        (
            "2022-11-14 13:21:15.341 Df My App[59972:780a15] (_ctypes.cpython-38-iphonesimulator.dylib) ",
            ("", True),
        ),
        # Unknown content
        (
            "This doesn't match the regex",
            ("This doesn't match the regex", False),
        ),
        # Log content that contains square brackets
        (
            "2022-11-14 13:21:15.341 Df My App[59972:780a15] (libffi.dylib) Test [1/5] ... OK",
            ("Test [1/5] ... OK", True),
        ),
        # Log content that contains `.so`
        (
            "2022-11-14 13:21:15.341 Df My App[59972:780a15] (_ctypes.cpython-312-iphonesimulator.so) "
            "A problem (foo.so) try to avoid it",
            ("A problem (foo.so) try to avoid it", True),
        ),
        # Log content that contains `.dylib`
        (
            "2022-11-14 13:21:15.341 Df My App[59972:780a15] (_ctypes.cpython-312-iphonesimulator.dylib) "
            "A problem (foo.dylib) try to avoid it",
            ("A problem (foo.dylib) try to avoid it", True),
        ),
    ],
)
def test_filter(original, filtered):
    assert macOS_log_clean_filter(original) == filtered



from unittest.mock import MagicMock

import pytest

import briefcase.platforms.linux.flatpak
from briefcase.console import Console, Log
from briefcase.exceptions import BriefcaseConfigError
from briefcase.integrations.flatpak import Flatpak
from briefcase.integrations.subprocess import Subprocess
from briefcase.platforms.linux.flatpak import LinuxFlatpakCreateCommand


@pytest.fixture
def create_command(tmp_path):
    return LinuxFlatpakCreateCommand(
        logger=Log(),
        console=Console(),
        base_path=tmp_path / "base_path",
        data_path=tmp_path / "briefcase",
    )


def test_binary_path(create_command, first_app_config, tmp_path):
    """The binary path is the marker file."""
    binary_path = create_command.binary_path(first_app_config)

    expected_path = (
        tmp_path
        / "base_path"
        / "build"
        / "first-app"
        / "linux"
        / "flatpak"
        / "com.example.first-app"
    )
    assert binary_path == expected_path


def test_project_path(create_command, first_app_config, tmp_path):
    """The project path is the bundle path."""
    project_path = create_command.project_path(first_app_config)
    bundle_path = create_command.bundle_path(first_app_config)

    expected_path = tmp_path / "base_path/build/first-app/linux/flatpak"
    assert expected_path == project_path == bundle_path


def test_distribution_path(create_command, first_app_config, tmp_path):
    """The distribution path is a flatpak bundle."""
    # Force the architecture to something odd for test purposes.
    create_command.tools.host_arch = "gothic"
    distribution_path = create_command.distribution_path(first_app_config)

    expected_path = tmp_path / "base_path/dist/First_App-0.0.1-gothic.flatpak"
    assert distribution_path == expected_path


def test_default_runtime_repo(create_command, first_app_config):
    """The flathub repository is the default runtime repository."""
    expected_repo = ("flathub", "https://flathub.org/repo/flathub.flatpakrepo")
    assert create_command.flatpak_runtime_repo(first_app_config) == expected_repo


def test_custom_runtime_repo(create_command, first_app_config):
    """A custom runtime repo can be specified."""
    first_app_config.flatpak_runtime_repo_alias = "custom_repo"
    first_app_config.flatpak_runtime_repo_url = "https://example.com/flatpak/runtime"

    expected_repo = ("custom_repo", "https://example.com/flatpak/runtime")
    assert create_command.flatpak_runtime_repo(first_app_config) == expected_repo


def test_custom_runtime_repo_no_alias(create_command, first_app_config):
    """If a custom runtime repo URL is specified, an alias must be specified as well."""
    first_app_config.flatpak_runtime_repo_url = "https://example.com/flatpak/runtime"

    with pytest.raises(
        BriefcaseConfigError,
        match="If you specify a custom Flatpak runtime repository",
    ):
        create_command.flatpak_runtime_repo(first_app_config)


def test_custom_runtime(create_command, first_app_config):
    """A custom runtime can be specified."""
    first_app_config.flatpak_runtime = "org.beeware.Platform"
    first_app_config.flatpak_runtime_version = "37.42"
    first_app_config.flatpak_sdk = "org.beeware.SDK"

    assert create_command.flatpak_runtime(first_app_config) == "org.beeware.Platform"
    assert create_command.flatpak_runtime_version(first_app_config) == "37.42"
    assert create_command.flatpak_sdk(first_app_config) == "org.beeware.SDK"


def test_missing_runtime(create_command, first_app_config):
    """Error if App config is missing the runtime."""
    first_app_config.flatpak_runtime_version = "37.42"
    first_app_config.flatpak_sdk = "org.beeware.SDK"

    with pytest.raises(
        BriefcaseConfigError,
        match="Briefcase configuration error: The App does not specify the Flatpak runtime to use",
    ):
        create_command.flatpak_runtime(first_app_config)


def test_missing_sdk(create_command, first_app_config):
    """Error if App config is missing the SDK."""
    first_app_config.flatpak_runtime = "org.beeware.Platform"
    first_app_config.flatpak_runtime_version = "37.42"

    with pytest.raises(
        BriefcaseConfigError,
        match="Briefcase configuration error: The App does not specify the Flatpak SDK to use",
    ):
        create_command.flatpak_sdk(first_app_config)


def test_missing_runtime_version(create_command, first_app_config):
    """Error if App config is missing the runtime version."""
    first_app_config.flatpak_runtime = "org.beeware.Platform"
    first_app_config.flatpak_sdk = "org.beeware.SDK"

    with pytest.raises(
        BriefcaseConfigError,
        match="Briefcase configuration error: The App does not specify the version of the Flatpak runtime to use",
    ):
        create_command.flatpak_runtime_version(first_app_config)


def test_verify_linux(create_command, monkeypatch, tmp_path):
    """Verifying on Linux creates an SDK wrapper."""
    create_command.tools.host_os = "Linux"
    create_command.tools.subprocess = MagicMock(spec_set=Subprocess)

    mock_flatpak_verify = MagicMock(wraps=Flatpak.verify)
    monkeypatch.setattr(
        briefcase.platforms.linux.flatpak.Flatpak,
        "verify",
        mock_flatpak_verify,
    )

    # Verify the tools
    create_command.verify_tools()

    # Flatpak tool was verified
    mock_flatpak_verify.assert_called_once_with(tools=create_command.tools)
    assert isinstance(create_command.tools.flatpak, Flatpak)



from unittest import mock

import pytest

from briefcase.commands.run import LogFilter
from briefcase.integrations.subprocess import StopStreaming


def test_default_filter():
    """A default logfilter echoes content verbatim."""
    popen = mock.MagicMock()
    log_filter = LogFilter(
        popen,
        clean_filter=None,
        clean_output=True,
        exit_filter=None,
    )

    for i in range(0, 10):
        line = f"this is line {i}"

        # Every line is returned verbatim
        assert [line] == list(log_filter(line))

    # no return code was detected
    assert log_filter.returncode is None


def test_clean_filter():
    """A cleaning filter can be used to strip content."""

    # Define a cleaning filter that strips the first 5 characters,
    # and identifies all content as Python
    def clean_filter(line):
        return line[5:], True

    popen = mock.MagicMock()
    log_filter = LogFilter(
        popen,
        clean_filter=clean_filter,
        clean_output=True,
        exit_filter=None,
    )

    for i in range(0, 10):
        line = f"this is line {i}"

        # The line has the preamble stripped
        assert [line[5:]] == list(log_filter(line))

    # no return code was detected
    assert log_filter.returncode is None


def test_clean_filter_unclean_output():
    """A cleaning filter can be used to strip content, but doesn't have to alter
    output."""

    # Define a cleaning filter that strips the first 5 characters,
    # and identifies all content as Python
    def clean_filter(line):
        return line[5:], True

    popen = mock.MagicMock()
    log_filter = LogFilter(
        popen,
        clean_filter=clean_filter,
        clean_output=False,
        exit_filter=None,
    )

    for i in range(0, 10):
        line = f"this is line {i}"

        # Every line is returned verbatim
        assert [line] == list(log_filter(line))

    # no return code was detected
    assert log_filter.returncode is None


@pytest.mark.parametrize(
    "raw, expected_output, use_content_filter, clean_output, returncode",
    [
        # Without cleaning, simple content is passed through as is.
        (
            ["1: line 1", "2: line 2", "3: line 3", "4: line 4", "5: line 5"],
            ["1: line 1", "2: line 2", "3: line 3", "4: line 4", "5: line 5"],
            False,  # don't use content filter
            False,  # don't use clean output
            None,  # no return code
        ),
        # Cleaning can be turned on without altering output
        (
            ["1: line 1", "2: line 2", "3: line 3", "4: line 4", "5: line 5"],
            ["1: line 1", "2: line 2", "3: line 3", "4: line 4", "5: line 5"],
            True,  # use content filter
            False,  # don't use clean output
            None,  # no return code
        ),
        # Dumped content is ignored, even if output isn't being cleaned
        (
            ["1: line 1", "2: line 2", "DUMP: garbage", "4: line 4", "5: line 5"],
            ["1: line 1", "2: line 2", "4: line 4", "5: line 5"],
            True,  # use content filter
            False,  # don't use clean output
            None,  # no return code
        ),
        # Output can be cleaned
        (
            ["1: line 1", "2: line 2", "3: line 3", "4: line 4", "5: line 5"],
            ["line 1", "line 2", "line 3", "line 4", "line 5"],
            True,  # use content filter
            True,  # use clean output
            None,  # no return code
        ),
        # Dumped content won't appear in cleaned output
        (
            ["1: line 1", "2: line 2", "DUMP: garbage", "4: line 4", "5: line 5"],
            ["line 1", "line 2", "line 4", "line 5"],
            True,  # use content filter
            True,  # use clean output
            None,  # no return code
        ),
        # Without cleaning, exit status won't be found
        (
            ["1: line 1", "2: line 2", "3: -----", "4: ", "5: EXIT 42", "post"],
            ["1: line 1", "2: line 2", "3: -----", "4: ", "5: EXIT 42", "post"],
            False,  # don't use content filter
            False,  # don't use clean output
            None,  # no return code due to line prefixes
        ),
        # If test output is clean without filtering, exit status can be determined
        (
            ["line 1", "line 2", "-----", "", "EXIT 42", "post"],
            ["line 1", "line 2", "-----", ""],
            False,  # don't use content filter
            False,  # don't use clean output
            42,  # exit status from raw output
        ),
        # Exit status can be found without altering output
        # Line prefixes are all even to ensure they aren't ignored
        (
            ["2: line 1", "4: line 2", "6: -----", "8: ", "10: EXIT 42", "post"],
            ["2: line 1", "4: line 2", "6: -----", "8: "],
            True,  # use content filter
            False,  # don't use clean output
            42,  # exit status
        ),
        # Exit status won't consider ignored content
        # Dumped content won't be considered
        (
            ["2: line 1", "4: -----", "DUMP: garbage", "6: ", "8: EXIT 42", "post"],
            ["2: line 1", "4: -----", "6: "],
            True,  # use content filter
            False,  # don't use clean output
            42,  # exit status
        ),
        # Exit status won't consider ignored content
        # Line 5 will be ignored, but not dumped
        (
            ["2: line 1", "4: -----", "5: Ignore this", "6: ", "8: EXIT 42", "post"],
            ["2: line 1", "4: -----", "5: Ignore this", "6: "],
            True,  # use content filter
            False,  # don't use clean output
            42,  # exit status
        ),
        # Exit status can be found with cleaned output
        # Line prefixes are all even to ensure they aren't ignored; output is clean
        (
            ["2: line 1", "4: line 2", "6: -----", "8: ", "10: EXIT 42", "post"],
            ["line 1", "line 2", "-----", ""],
            True,  # use content filter
            True,  # use clean output
            42,  # Exit status
        ),
        # Exit status won't consider ignored content
        # Dumped content won't be considered; output is clean
        (
            ["2: line 1", "4: -----", "DUMP: garbage", "6: ", "8: EXIT 42", "post"],
            ["line 1", "-----", ""],
            True,  # use content filter
            True,  # use clean output
            42,  # exit status
        ),
        # Exit status won't consider ignored content
        # Line 5 will be ignored, but not dumped; output is clean
        (
            ["2: line 1", "4: -----", "5: Ignore this", "6: ", "8: EXIT 42", "post"],
            ["line 1", "-----", "Ignore this", ""],
            True,  # use content filter
            True,  # use clean output
            42,  # exit status
        ),
        # Exit status won't be found if it occurs on ignored lines
        # Line 5 is the right pattern, but it will be marked as ignored by the cleaner
        (
            ["2: line 1", "4: line 2", "5: -----", "6: ", "8: EXIT 42", "post"],
            ["2: line 1", "4: line 2", "5: -----", "6: ", "8: EXIT 42", "post"],
            True,  # use content filter
            False,  # don't use clean output
            None,  # no return code due to line 5 not matching
        ),
    ],
)
def test_log_filter(
    raw,
    expected_output,
    use_content_filter,
    clean_output,
    returncode,
):
    """The log filter behaves as expected."""

    # Define a clean filter that removes an index at the start of the line, only
    # analyses content with an even prefix, and dumps content that starts "DUMP:"
    def clean_filter(line):
        if line.startswith("DUMP: "):
            return None
        parts = line.split(":", 1)
        try:
            line = int(parts[0])
            return " ".join(parts[1:]).strip(), line % 2 == 0
        except ValueError:
            return line, False

    # Define a custom filter that looks for specific multiline output
    exit_filter = LogFilter.test_filter(r"^-----\n\nEXIT (?P<returncode>\d+)$")

    # Set up a log stream
    popen = mock.MagicMock()
    log_filter = LogFilter(
        popen,
        clean_filter=clean_filter if use_content_filter else None,
        clean_output=clean_output,
        exit_filter=exit_filter,
    )

    # Pipe the raw output through the log filter, and capture the output
    output = []
    terminated = False
    for raw_line in raw:
        try:
            for line in log_filter(raw_line):
                output.append(line)
        except StopStreaming:
            terminated = True
            break

    # Actual output is as expected
    assert output == expected_output

    if returncode is None:
        # No success/failure condition was set; no termination condition was processed
        assert log_filter.returncode is None
        assert not terminated
    else:
        # The success/failure condition was detected, and the termination condition was processed
        assert log_filter.returncode == returncode
        assert terminated
